from ewah.hooks.base import EWAHBaseHook
from ewah.utils.run_commands import run_cmd
from airflow.utils.file import TemporaryDirectory
from tempfile import NamedTemporaryFile
from typing import Optional, Dict, Any

import os
import re


class EWAHRemoteGitHook(EWAHBaseHook):
    _ATTR_RELABEL = {
        "git_link": "host",
        "branch": "schema",
    }

    conn_name_attr = "ewah_git_conn_id"
    default_conn_name = "ewah_git_default"
    conn_type = "ewah_git"
    hook_name = "EWAH Git Remote Connection"

    @staticmethod
    def get_connection_form_widgets() -> Dict[str, Any]:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from ewah.utils.widgets import EWAHTextAreaWidget

        return {
            "extra__ewah_git__private_key": StringField(
                "Private SSH Key (RSA Format)",
                widget=EWAHTextAreaWidget(rows=12),
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        return {
            "hidden_fields": ["extra", "port", "login"],
            "relabeling": {
                "host": "Git SSH Link",
                "schema": "Branch (optional)",
                "password": "SSH Key Password (optional)",
            },
        }

    def clone_repo(self, clone_to: str, env: Optional[Dict[str, str]] = None) -> None:
        """Clone the repository into a specific location.

        :param clone_to: Directory to clone to.
        :param env: Dictionary of environment variables to use. Defaults to
            os.environ.copy(). Requires a $HOME environment variable to work with SSH.
        """

        env = env or os.environ.copy()

        cmd = []
        cmd.append("eval `ssh-agent`")
        with TemporaryDirectory(prefix="__ewah_git_") as tmp_dir:
            # temporarily save SSH key, if applicable, in this folder
            with NamedTemporaryFile(dir=tmp_dir) as ssh_key_file:
                if self.conn.private_key:
                    ssh_key_filepath = os.path.abspath(ssh_key_file.name)
                    ssh_key_file.write(self.conn.private_key.encode())
                    ssh_key_file.seek(0)
                    if self.conn.password:
                        passfile_path = os.path.abspath(
                            tmp_dir + os.path.sep + "ssh_key_pw"
                        )
                        with open(passfile_path, "w+") as passfile:
                            passfile.write(
                                '#!/bin/bash\necho "{0}"'.format(self.conn.password)
                            )
                            passfile.seek(0)
                        cmd.append("chmod 777 {0}".format(passfile_path))
                        cmd.append(
                            'DISPLAY=":0.0" SSH_ASKPASS="{0}" ssh-add {1}'.format(
                                passfile_path, ssh_key_filepath
                            )
                        )
                    else:
                        cmd.append("ssh-add {0}".format(ssh_key_filepath))

                # Add actual clone commands
                git_link = self.conn.git_link
                cmd.append("mkdir -p $HOME/.ssh")

                # Must add the host to the known hosts or it will ask for confirmation
                ssh_domain = re.search("@(.*):", git_link).group(1)
                cmd.append(
                    "ssh-keyscan -H {0} >> $HOME/.ssh/known_hosts".format(ssh_domain)
                )
                if self.conn.branch:
                    cmd.append(
                        "git clone -b {0} {1} {2}".format(
                            self.conn.branch, git_link, clone_to
                        )
                    )
                else:
                    cmd.append("git clone {0} {1}".format(git_link, clone_to))

                # Execute commands!
                assert run_cmd(cmd, env, self.log.info) == 0



import io
import requests
from msal import ConfidentialClientApplication
from openpyxl import load_workbook
from urllib.parse import quote

from ewah.hooks.base import EWAHBaseHook


class EWAHSharepointHook(EWAHBaseHook):
    GRAPH_API_BASE = "https://graph.microsoft.com"

    conn_name_attr: str = "ewah_sharepoint_conn_id"
    default_conn_name: str = "ewah_sharepoint_default"
    conn_type: str = "ewah_sharepoint"
    hook_name: str = "EWAH Microsoft Sharepoint Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["port", "extra", "host"],
            "relabeling": {
                "login": "Client ID",
                "password": "Client Secret",
                "schema": "Tenant ID",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget

        return {
            "extra__ewah_sharepoint__site_hostname": StringField(
                "Site Hostname (e.g. company.sharepoint.com)",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_sharepoint__site_path": StringField(
                "Site Path (e.g. sites/Finance)",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_sharepoint__drive_id": StringField(
                "Drive ID (optional, e.g. b!oS9JMAuH1...DdrP)",
                widget=BS3TextFieldWidget(),
            ),
        }

    def _authenticate_with_graph_api(self):
        """Authenticate with Microsoft Graph API using client credentials flow."""
        self.log.info("Authenticating with Microsoft Graph API...")

        # Get credentials from connection
        client_id = self.conn.login
        client_secret = self.conn.password
        tenant_id = self.conn.schema

        if not all([client_id, client_secret, tenant_id]):
            raise ValueError(
                "client_id, client_secret, and tenant_id must be configured in the connection"
            )

        authority = f"https://login.microsoftonline.com/{tenant_id}"
        scopes = ["https://graph.microsoft.com/.default"]

        try:
            app = ConfidentialClientApplication(
                client_id, authority=authority, client_credential=client_secret
            )

            result = app.acquire_token_for_client(scopes=scopes)

            if "access_token" in result:
                self.log.info("Successfully authenticated with Microsoft Graph API")
                return result["access_token"]
            else:
                error_msg = f"Authentication failed: {result.get('error')} - {result.get('error_description')}"
                self.log.error(error_msg)
                raise Exception(error_msg)
        except Exception as e:
            self.log.error(f"Authentication error: {str(e)}")
            raise

    def _get_site_id(self, access_token, site_hostname, site_path):
        """Get the SharePoint site ID from the site hostname and path."""
        self.log.info(f"Getting site ID for {site_hostname}/{site_path}")
        site_url = f"{self.GRAPH_API_BASE}/v1.0/sites/{site_hostname}:/{site_path}"
        site_resp = requests.get(
            site_url, headers={"Authorization": f"Bearer {access_token}"}
        )
        site_resp.raise_for_status()
        site_id = site_resp.json()["id"]
        self.log.info(f"Retrieved site ID: {site_id}")
        return site_id

    def _get_drive_id(self, access_token, site_id):
        """Get the default drive ID from the SharePoint site."""
        self.log.info(f"Getting drive ID for site {site_id}")
        drive_url = f"{self.GRAPH_API_BASE}/v1.0/sites/{site_id}/drives"
        drive_resp = requests.get(
            drive_url, headers={"Authorization": f"Bearer {access_token}"}
        )
        drive_resp.raise_for_status()
        drive_id = drive_resp.json()["value"][0]["id"]
        self.log.info(f"Retrieved drive ID: {drive_id}")
        return drive_id

    def get_data_from_excel(
        self,
        relative_file_path: str,
        worksheet_name: str,
        header_row: int,
        start_row: int,
        batch_size: int,
        site_hostname: str = None,
        site_path: str = None,
        drive_id: str = None,
    ):
        """
        Get data from Excel file in SharePoint using Microsoft Graph API.

        Args:
            relative_file_path: Path to the Excel file relative to the SharePoint site
            worksheet_name: Name of the worksheet to read
            header_row: Row number containing headers (1-based)
            start_row: Row number to start reading data from (1-based)
            batch_size: Number of rows to process in each batch
            site_hostname: SharePoint hostname (e.g., "company.sharepoint.com")
            site_path: SharePoint site path (e.g., "sites/MySite")
            drive_id: SharePoint drive ID
        """
        assert (
            isinstance(batch_size, int) and batch_size > 0
        ), f"type: {type(batch_size)}, value: {str(batch_size)}"
        assert (
            isinstance(header_row, int) and header_row > 0
        ), f"type: {type(header_row)}, value: {str(header_row)}"
        assert (
            isinstance(start_row, int) and start_row > 0
        ), f"type: {type(start_row)}, value: {str(start_row)}"
        assert start_row > header_row

        # Get site_hostname and site_path from connection if not provided
        if not site_hostname:
            site_hostname = self.conn.extra_dejson.get(
                "extra__ewah_sharepoint__site_hostname"
            )
        if not site_path:
            site_path = self.conn.extra_dejson.get("extra__ewah_sharepoint__site_path")

        if not drive_id:
            drive_id = self.conn.extra_dejson.get("extra__ewah_sharepoint__drive_id")

        if not site_hostname or not site_path:
            raise ValueError(
                "site_hostname and site_path must be provided either as parameters or in the connection"
            )

        # Authenticate with Microsoft Graph API
        access_token = self._authenticate_with_graph_api()

        # Get site ID
        site_id = self._get_site_id(access_token, site_hostname, site_path)

        # If drive_id is not provided, get it using the default method
        if not drive_id:
            drive_id = self._get_drive_id(access_token, site_id)

        # Download the file using Graph API
        self.log.info(f"Downloading file: {relative_file_path}")
        file_path_encoded = quote(relative_file_path.strip().strip("/"))
        download_url = f"{self.GRAPH_API_BASE}/v1.0/sites/{site_id}/drives/{drive_id}/root:/{file_path_encoded}:/content"

        file_resp = requests.get(
            download_url, headers={"Authorization": f"Bearer {access_token}"}
        )
        file_resp.raise_for_status()

        # Process the file content
        bytes_file_obj = io.BytesIO()
        bytes_file_obj.write(file_resp.content)
        bytes_file_obj.seek(0)

        try:
            workbook = load_workbook(bytes_file_obj, data_only=True)
            if worksheet_name not in workbook.sheetnames:
                raise ValueError(
                    f"Worksheet '{worksheet_name}' not found. Available worksheets: {workbook.sheetnames}"
                )
            ws = workbook[worksheet_name]

            # Get headers
            headers = {
                ws.cell(row=header_row, column=col).value: col
                for col in range(1, ws.max_column + 1)
                if ws.cell(row=header_row, column=col).value is not None
            }

            if not headers:
                self.log.warning(f"No headers found in row {header_row}")
                return

            self.log.info(f"Found {len(headers)} columns: {list(headers.keys())}")

            # Process data in batches
            while start_row <= ws.max_row:
                self.log.info(
                    f"Yielding batch of {batch_size} rows starting from row {start_row}."
                )
                batch_data = []
                for row in range(
                    start_row, min(start_row + batch_size, ws.max_row + 1)
                ):
                    row_data = {}
                    for k, v in headers.items():
                        cell_value = ws.cell(row=row, column=v).value
                        row_data[k] = cell_value
                    batch_data.append(row_data)

                if batch_data:
                    yield batch_data
                start_row += batch_size

        except Exception as e:
            self.log.error(f"Error processing Excel file: {str(e)}")
            raise



from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.utils.airflow_utils import EWAHSqlSensor, datetime_utcnow_with_tz
from ewah.utils.dbt_operator import EWAHdbtOperator

from croniter import croniter
from datetime import datetime, timedelta
from typing import Optional, Union

import pytz


def dbt_dags_factory(
    airflow_conn_id,
    repo_type,
    dwh_engine,
    dwh_conn_id,
    database_name=None,
    git_conn_id=None,  # if provided, expecting private SSH key in conn extra
    local_path=None,
    dbt_version=None,  # Defaults to require-dbt-version in dbt_project.yml
    subfolder=None,  # optional: supply if dbt project is in a subfolder
    threads=4,  # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-threads
    schema_name="analytics",  # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-target-schemas
    keepalives_idle=0,  # see https://docs.getdbt.com/reference/warehouse-profiles/postgres-profile/
    dag_base_name="T_dbt_run",
    schedule_interval: Optional[Union[str, timedelta]] = timedelta(days=1),
    full_refresh_schedule_interval: Optional[Union[str, timedelta]] = None,
    start_date=datetime(2019, 1, 1),
    default_args=None,
    run_flags=None,  # e.g. --model tag:base
    seed_flags=None,
    project=None,  # BigQuery alias
    dataset=None,  # BigQuery alias
    dagrun_timeout_factor=None,  # doesn't apply to full refresh
    task_timeout_factor=0.8,  # doesn't apply to full refresh
    metabase_conn_id=None,  # push docs to Metabase after full refresh run if exists
    env_var_conn_ids=None,
):
    run_flags = run_flags or ""  # use empty string instead of None
    seed_flags = seed_flags or ""

    # only PostgreSQL & Snowflake implemented as of now!
    assert dwh_engine in (
        EC.DWH_ENGINE_POSTGRES,
        EC.DWH_ENGINE_SNOWFLAKE,
        EC.DWH_ENGINE_BIGQUERY,
    )

    # Initialize parameters with default values here that can be updated depending
    # on schedule_interval conditions
    catchup = False
    end_date = None
    # Keep original start_date for full refresh DAG
    full_refresh_start_date = start_date

    # Allow using cron-style schedule intervals
    if isinstance(schedule_interval, str):
        assert croniter.is_valid(
            schedule_interval
        ), "schedule_interval is not valid a cron expression!"
    elif isinstance(schedule_interval, timedelta):
        # if start_date is timezone offset-naive, assume utc and turn into offset-aware
        catchup = True
        if not start_date.tzinfo:
            start_date = start_date.replace(tzinfo=pytz.utc)

        start_date += (
            int((datetime_utcnow_with_tz() - start_date) / schedule_interval) - 1
        ) * schedule_interval
        end_date = start_date + 2 * schedule_interval - timedelta(seconds=1)

    dag_kwargs = {
        "catchup": catchup,
        "start_date": start_date,
        "end_date": end_date,
        "default_args": default_args,
        "max_active_runs": 1,
    }

    if dagrun_timeout_factor and isinstance(schedule_interval, timedelta):
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        dagrun_timeout = dagrun_timeout_factor * schedule_interval
    else:
        dagrun_timeout = None

    if task_timeout_factor and isinstance(schedule_interval, timedelta):
        _msg = "task_timeout_factor must be a number between 0 and 1!"
        assert isinstance(task_timeout_factor, (int, float)) and (
            0 < task_timeout_factor <= 1
        ), _msg
        execution_timeout = task_timeout_factor * schedule_interval
    else:
        execution_timeout = None

    dag_1 = DAG(
        dag_base_name,
        schedule_interval=schedule_interval,
        dagrun_timeout=dagrun_timeout,
        **dag_kwargs,
    )
    # for full refresh we need to set the original start_date
    dag_kwargs["start_date"] = full_refresh_start_date
    dag_kwargs["catchup"] = False # dont catch up full refreshes
    dag_2 = DAG(dag_base_name + "_full_refresh", schedule_interval=full_refresh_schedule_interval, **dag_kwargs)

    sensor_sql = """
        SELECT
            -- only succeed if there is no other running DagRun
            CASE WHEN COUNT(*) = 0 THEN 1 ELSE 0 END
        FROM public.dag_run
        WHERE dag_id IN ('{0}', '{1}')
          AND state = 'running'
          AND data_interval_end < '{2}' -- DagRun's data_interval_end
          AND NOT (run_id = '{3}' AND dag_id = '{4}')
          -- Note: data_interval_end = data_interval_start if run_type = 'manual'
    """.format(
        dag_1._dag_id,
        dag_2._dag_id,
        "{{ data_interval_end }}",
        "{{ run_id }}",
        "{{ dag._dag_id }}",
    )

    snsr_1 = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag_1,
    )
    snsr_2 = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag_2,
    )

    dbt_kwargs = {
        "repo_type": repo_type,
        "dwh_conn_id": dwh_conn_id,
        "git_conn_id": git_conn_id,
        "local_path": local_path,
        "dbt_version": dbt_version,
        "subfolder": subfolder,
        "threads": threads,
        "schema_name": schema_name,
        "keepalives_idle": 0,
        "dwh_engine": dwh_engine,
        "database_name": database_name,
        "project": project,
        "dataset": dataset,
        "env_var_conn_ids": env_var_conn_ids,
    }

    run_1 = EWAHdbtOperator(
        task_id="dbt_run",
        dbt_commands=[f"seed {seed_flags}", f"run {run_flags}"],
        dag=dag_1,
        execution_timeout=execution_timeout,
        **dbt_kwargs,
    )
    run_2 = EWAHdbtOperator(
        task_id="dbt_run",
        dbt_commands=[
            f"seed --full-refresh {seed_flags}",
            f"run --full-refresh {run_flags}",
        ],
        dag=dag_2,
        # If metabase_conn_id exists, push dbt docs to Metabase after full refresh run
        metabase_conn_id=metabase_conn_id,
        **dbt_kwargs,
    )

    run_flags_freshness = (
        run_flags.replace("--models", "--select")
        .replace("--model", "--select")
        .replace("-m ", "-s ")
    )
    test_1 = EWAHdbtOperator(
        task_id="dbt_test",
        dbt_commands=[f"test {run_flags}", f"source freshness {run_flags_freshness}"],
        dag=dag_1,
        execution_timeout=execution_timeout,
        **dbt_kwargs,
    )
    test_2 = EWAHdbtOperator(
        task_id="dbt_test", dbt_commands=f"test {run_flags}", dag=dag_2, **dbt_kwargs
    )

    snsr_1 >> run_1 >> test_1
    snsr_2 >> run_2 >> test_2

    return (dag_1, dag_2)


def dbt_snapshot_dag(
    dwh_engine,
    dwh_conn_id,
    repo_type="git",
    database_name=None,
    git_conn_id=None,
    local_path=None,
    dbt_version=None,
    subfolder=None,
    threads=4,
    schema_name="analytics",  # for the profiles.yml
    keepalives_idle=0,
    dag_name="T_dbt_snapshots",
    schedule_interval: Optional[Union[str, timedelta]] = timedelta(hours=1),
    start_date=None,
    default_args=None,
    dagrun_timeout_factor=None,
    task_timeout_factor=0.8,
    project=None,  # BigQuery alias
    dataset=None,  # BigQuery alias
    run_flags=None,
    env_var_conn_ids=None,
):
    run_flags = run_flags or ""

    # only PostgreSQL & Snowflake implemented as of now!
    assert dwh_engine in (
        EC.DWH_ENGINE_POSTGRES,
        EC.DWH_ENGINE_SNOWFLAKE,
        EC.DWH_ENGINE_BIGQUERY,
    )

    if dagrun_timeout_factor and isinstance(schedule_interval, timedelta):
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        dagrun_timeout = dagrun_timeout_factor * schedule_interval
    else:  # In case of 0 set to None
        dagrun_timeout = None

    if task_timeout_factor and isinstance(schedule_interval, timedelta):
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(task_timeout_factor, (int, float)) and (
            0 < task_timeout_factor <= 1
        ), _msg
        execution_timeout = task_timeout_factor * schedule_interval
    else:
        execution_timeout = None

    dag = DAG(
        dag_name,
        schedule_interval=schedule_interval,
        catchup=False,
        max_active_runs=1,
        start_date=start_date,
        default_args=default_args,
        dagrun_timeout=dagrun_timeout,
    )

    task = EWAHdbtOperator(
        dag=dag,
        task_id="dbt_snapshot",
        dbt_commands=[f"snapshot {run_flags}"],
        repo_type=repo_type,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        git_conn_id=git_conn_id,
        local_path=local_path,
        database_name=database_name,
        dbt_version=dbt_version,
        subfolder=subfolder,
        threads=threads,
        schema_name=schema_name,
        keepalives_idle=keepalives_idle,
        execution_timeout=execution_timeout,
        project=project,
        dataset=dataset,
        env_var_conn_ids=env_var_conn_ids,
    )

    return dag



from ewah.constants import EWAHConstants as EC
from ewah.hooks.plentymarkets import EWAHPlentyMarketsHook
from ewah.operators.base import EWAHBaseOperator

from datetime import datetime, date


class EWAHPlentyMarketsOperator(EWAHBaseOperator):
    _NAMES = ["plentymarkets"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHPlentyMarketsHook.conn_type

    def __init__(
        self,
        resource=None,
        additional_api_call_params=None,
        batch_size=10000,
        request_method="get",
        post_request_payload=None,
        expand_field=None, #name of the field to expand
        *args,
        **kwargs,
    ):
        kwargs["primary_key"] = kwargs.get("primary_key", "id")
        resource = resource or kwargs.get("target_table_name")
        
        # expand_field has only been implemented for full refresh strategy
        assert (
            kwargs["extract_strategy"] == EC.ES_FULL_REFRESH or expand_field is None
        ), "expand_field can only be used with full refresh strategy"
        
        if kwargs["extract_strategy"] == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get("subsequent_field", "updatedAt")
            assert (
                EWAHPlentyMarketsHook.format_resource(resource)
                in EWAHPlentyMarketsHook._INCREMENTAL_FIELDS.keys()
            ), "{0} is not subsequent loadable!".format(resource)
        if kwargs["extract_strategy"] == EC.ES_INCREMENTAL:
            assert (
                EWAHPlentyMarketsHook.format_resource(resource)
                in EWAHPlentyMarketsHook._INCREMENTAL_FIELDS.keys()
            ), "{0} is not incrementally loadable!".format(resource)
        super().__init__(*args, **kwargs)

        assert isinstance(additional_api_call_params, (type(None), dict))
        self.resource = resource
        self.additional_api_call_params = additional_api_call_params
        self.batch_size = batch_size
        self.request_method = request_method
        self.post_request_payload = post_request_payload
        self.expand_field = expand_field

    # method to expand a field that contains an array of dictionaries
    def expand_field_data(self, batch):
        if not self.expand_field:
            return batch

        expanded_data = []

        while batch:
            row = batch.pop(0)

            # skip if field doesn't exist in this row and just append row as is
            if self.expand_field not in row:
                expanded_data.append(row)
                continue
            
            # get the field data
            field_data = row.pop(self.expand_field, [])

            # If field exists but is None/empty, default to empty list
            if not field_data:
                expanded_data.append(row)
                continue

            # Create a new row for each item in the field
            for item in field_data:
                new_row = row.copy()
                # Create new columns with prefixed names to avoid conflicts
                prefixed_item = {f"{self.expand_field}_{k}": v for k, v in item.items()}
                new_row.update(prefixed_item)
                expanded_data.append(new_row)

        return expanded_data

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            # This is likely to be a text field - need a date instead
            if not isinstance(data_from, date):
                if isinstance(data_from, datetime):
                    data_from = data_from.date()
                elif isinstance(data_from, str):
                    data_from = datetime.fromisoformat(data_from).date()
                else:
                    raise Exception(
                        "Data type of {0} is invalid!".format(self.subsequent_field)
                    )
        else:
            data_from = self.data_from

        total_records = 0
        for batch in self.source_hook.get_data_in_batches(
            resource=self.resource,
            data_from=data_from,
            data_until=self.data_until,
            additional_params=self.additional_api_call_params,
            batch_size=self.batch_size,
            request_method=self.request_method,
            post_request_payload=self.post_request_payload,
        ):
            batch_size = len(batch)
            self.log.info(f"Processing batch of {batch_size} records")

            if self.expand_field:
                self.log.info(f"Expanding fields: {self.expand_field}")
            batch = self.expand_field_data(batch)

            expanded_size = len(batch)
            if expanded_size != batch_size:
                self.log.info(f"Batch size after expansion: {expanded_size} records")

            self.upload_data(batch)
            total_records += expanded_size

        self.log.info(
            f"Completed extraction of {total_records} total records for {self.resource}"
        )



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.providers.amazon.aws.hooks.s3 import S3Hook

from botocore.exceptions import ClientError
from datetime import datetime, timedelta
from io import StringIO
from multiprocessing.dummy import Pool as ThreadPool

import json
import csv
import gzip
import time


class ExtendedS3Hook(S3Hook):
    """Extends and improves the provider package's S3Hook's capability"""

    credentials_by_region = {}
    force_refresh = False
    is_refreshing = {}

    def _get_credentials(self, region_name, force_refresh=False):
        "Avoid creation of a new session each time an S3 file is downloaded."
        if (not self.is_refreshing.get(region_name)) and (
            force_refresh
            or self.force_refresh
            or (not region_name in self.credentials_by_region)
        ):
            # Avoid all threads renewing credentials at the same time
            self.is_refreshing[region_name] = True
            self.force_refresh = False
            self.credentials_by_region[region_name] = super()._get_credentials(
                region_name=region_name
            )
            self.is_refreshing[region_name] = False
        while self.is_refreshing.get(region_name):
            # Wait until the appropriate thread is finished refreshing credentials
            time.sleep(0.1)
        return self.credentials_by_region[region_name]

    def read_key(self, key, bucket_name=None, is_compressed=False):
        """
        Overwrite method in order to allow for compressed objects
        Source: https://airflow.apache.org/docs/apache-airflow/1.10.12/_api/airflow/hooks/S3_hook/index.html#airflow.hooks.S3_hook.S3Hook.read_key
        """
        obj = self.get_key(key, bucket_name)
        if is_compressed:
            return obj.get()["Body"].read()
        return obj.get()["Body"].read().decode("utf-8")


class EWAHS3Operator(EWAHBaseOperator):
    """Only implemented for JSON and CSV files from S3 right now!"""

    _NAMES = ["s3"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _IMPLEMENTED_FORMATS = [
        "JSON",
        "JSONL",  # Multiple JSON objects split by newlines in the same file
        "AWS_FIREHOSE_JSON",
        "CSV",
    ]

    _BOM = b"\xef\xbb\xbf"

    def __init__(
        self,
        bucket_name,
        file_format,
        prefix="",  # use for subfolder structures
        suffix=None,  # use e.g. for file types
        key_name=None,  # if not specified, uses data_from and data_until
        csv_format_options={},
        csv_encoding="utf-8",
        decompress=False,
        file_load_parallelism=1000,
        thread_pool_size=20,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        if not file_format in self._IMPLEMENTED_FORMATS:
            raise Exception(
                "File format not implemented! Available formats:"
                + " {0}".format(str(self._IMPLEMENTED_FORMATS))
            )

        if key_name:
            if self.load_data_from or self.load_data_until:
                raise Exception(
                    "Both key_name and either data_from or data_"
                    + "until specified! Cannot have both specified."
                )

        if not file_format == "CSV" and csv_format_options:
            raise Exception("csv_format_options is only valid for CSV files!")

        self.bucket_name = bucket_name
        self.prefix = prefix
        self.suffix = suffix
        self.key_name = key_name
        self.file_format = file_format
        self.csv_format_options = csv_format_options
        self.csv_encoding = csv_encoding
        self.decompress = decompress
        self.file_load_parallelism = file_load_parallelism
        self.thread_pool_size = thread_pool_size

    def _iterate_through_bucket(
        self,
        s3hook,
        bucket,
        prefix,
        modified_from=None,
        modified_until=None,
        suffix=None,
        file_load_parallelism=1000,
        thread_pool_size=20,
    ):
        """The bucket.objects.filter() method only returns a max of 1000
        objects. If more objects are in an S3 bucket, pagniation is
        required. See also: https://stackoverflow.com/questions/44238525/how-to-iterate-over-files-in-an-s3-bucket
        """

        def read_key(key_name):
            return s3hook.read_key(key_name, bucket, self.decompress)

        cli = s3hook.get_client_type("s3")
        paginator = cli.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

        all_objects = [
            o
            for o in page_iterator.build_full_result()["Contents"]
            if (not modified_from or o["LastModified"] > modified_from)
            and (not modified_until or o["LastModified"] <= modified_until)
            and (not suffix or suffix == o["Key"][-len(suffix) :])
        ]

        self._temp_object_data = {}

        list_of_keys = [o["Key"] for o in all_objects]
        self.log.info("Iterating through {0} objects..".format(len(all_objects)))

        for item in all_objects:
            if not item["Key"] in self._temp_object_data:
                # use multithreading to efficiently load s3 files
                self.log.info(
                    "Getting next {0} files...".format(str(file_load_parallelism))
                )
                _next_keys = list_of_keys[:file_load_parallelism]
                del list_of_keys[:file_load_parallelism]
                pool = ThreadPool(thread_pool_size)
                try:
                    results = pool.map(read_key, _next_keys)
                except ClientError:
                    # This error can occur when the aws token expires - try refreshing
                    # the connection and see if the error persists
                    s3hook.force_refresh = True
                    results = pool.map(read_key, _next_keys)
                pool.close()
                pool.join()
                self._temp_object_data.update(dict(zip(_next_keys, results)))
            item["_body"] = self._temp_object_data.pop(item["Key"])
            yield item

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            self.data_from = self.get_max_value_of_column(self.subsequent_field)
        if self.file_format == "JSON":
            return self.execute_json(
                context=context,
                f_get_data=lambda file_content: json.loads(file_content),
            )
        elif self.file_format == "JSONL":
            f_get_data = lambda file_content: list(
                # Use StringIO in order to iterate through each line of the file
                map(json.loads, StringIO(file_content))
            )
            if self.decompress:
                f_get_data = lambda file_content: list(
                    map(
                        json.loads,
                        # We need to decompress the whole file before iterating over its lines
                        StringIO(gzip.decompress(bytes(file_content)).decode("utf-8")),
                    )
                )
            return self.execute_json(
                context=context,
                f_get_data=f_get_data,
            )
        elif self.file_format == "AWS_FIREHOSE_JSON":
            return self.execute_json(
                context=context,
                f_get_data=lambda file_content: json.loads(
                    "["  # hotfix for a weird encoding issue
                    + file_content.replace("\x00", "")
                    .replace("\u0000", "")
                    .replace("\\u0000", "")
                    .replace("\\x00", "")
                    .replace("}{", "},{")
                    + "]"
                ),
            )
        elif self.file_format == "CSV":
            return self.execute_csv(
                context=context,
            )
        else:
            raise Exception("File format not implemented!")

    def execute_csv(self, context):
        def chunk_upload(csv_reader):
            i = 0
            data = []
            for row in csv_reader:
                i += 1
                data.append(row)
                if i >= 100000:
                    self.upload_data(data=data)
                    data = []
                    i = 0
            if data:
                self.upload_data(data=data)

        hook = ExtendedS3Hook(self.source_conn.conn_id)
        if self.key_name:
            obj = hook.get_key(self.key_name, self.bucket_name)
            raw_data = obj.get()["Body"].read()
            if self.decompress:
                raw_data = gzip.decompress(raw_data)
            if raw_data[:3] == self._BOM:
                raw_data = raw_data[3:]
                csv_encoding = "utf-8"
            else:
                csv_encoding = self.csv_encoding
            try:
                raw_data = raw_data.decode(csv_encoding).splitlines()
            except UnicodeDecodeError:
                if csv_encoding == self.csv_encoding:
                    raise
                raw_data = raw_data.decode(self.csv_encoding).splitlines()
            reader = csv.DictReader(raw_data, **self.csv_format_options)
            chunk_upload(reader)

        else:
            for obj_iter in self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
                modified_from=self.data_from,
                modified_until=self.data_until,
                suffix=self.suffix,
                file_load_parallelism=self.file_load_parallelism,
                thread_pool_size=self.thread_pool_size,
            ):
                self.log.info("Loading data from file {0}".format(obj_iter["Key"]))
                raw_data = obj_iter["_body"]
                if self.decompress:
                    raw_data = gzip.decompress(raw_data)
                # remove BOM if it exists
                # also, if file has a BOM, it is 99.9% utf-8 encoded!
                if raw_data[:3] == self._BOM:
                    raw_data = raw_data[3:]
                    csv_encoding = "utf-8"
                else:
                    csv_encoding = self.csv_encoding
                # there may be a BOM while still not utf-8 -> use the
                # given csv_encoding argument if so
                try:
                    raw_data = raw_data.decode(csv_encoding).splitlines()
                except UnicodeDecodeError:
                    if csv_encoding == self.csv_encoding:
                        raise
                    raw_data = raw_data.decode(self.csv_encoding).splitlines()
                reader = csv.DictReader(raw_data, **self.csv_format_options)
                if self.subsequent_field:
                    # TODO: Abstract this logic
                    for datum in reader:
                        datum[self.subsequent_field] = datetime.strptime(
                            datum[self.subsequent_field], "%Y-%m-%dT%H:%M:%S.%f%z"
                        )
                self._metadata.update(
                    {
                        "bucket_name": self.bucket_name,
                        "file_name": obj_iter["Key"],
                        "file_last_modified": str(obj_iter["LastModified"]),
                    }
                )
                chunk_upload(reader)

    def execute_json(self, context, f_get_data):
        hook = ExtendedS3Hook(self.source_conn.conn_id)

        if self.key_name:
            data = f_get_data(
                file_content=hook.read_key(key=self.key_name, bucket=self.bucket_name)
            )
            self.upload_data(data=data)
        else:
            data = []
            objects = self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
                modified_from=self.data_from,
                modified_until=self.data_until,
                suffix=self.suffix,
                file_load_parallelism=self.file_load_parallelism,
                thread_pool_size=self.thread_pool_size,
            )
            for obj_iter in objects:
                self.log.info(
                    "Loading data from file {0}".format(
                        obj_iter["Key"],
                    )
                )
                self._metadata.update(
                    {
                        "bucket_name": self.bucket_name,
                        "file_name": obj_iter["Key"],
                        "file_last_modified": obj_iter["LastModified"],
                    }
                )
                data = f_get_data(file_content=obj_iter["_body"])
                if self.subsequent_field:
                    # TODO: Abstract this logic
                    for datum in data:
                        # Avoid KeyError when using `file_last_modified` as subsequent field
                        if not self.subsequent_field in self._metadata.keys():
                            datum[self.subsequent_field] = datetime.strptime(
                                datum[self.subsequent_field], "%Y-%m-%dT%H:%M:%S.%f%z"
                            )
                self.upload_data(data=data)



from ewah.constants import EWAHConstants as EC
from ewah.hooks.airflow import EWAHAirflowHook
from ewah.operators.base import EWAHBaseOperator

from dateutil.parser import parse


class EWAHAirflowOperator(EWAHBaseOperator):
    _NAMES = ["airflow"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: True,  # Only for DagRuns!
    }

    _CONN_TYPE = EWAHAirflowHook.conn_type

    def __init__(self, endpoint=None, request_page_size=100, *args, **kwargs):
        self.endpoint = endpoint or kwargs.get("target_table_name")
        self.request_page_size = request_page_size
        assert not kwargs.get("subsequent_field"), "Not allowed for this operator!"
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            # Only for DagRuns!
            assert (
                EWAHAirflowHook._ENDPOINTS.get(self.endpoint, self.endpoint)
                == EWAHAirflowHook._ENDPOINT_DAGRUNS
            ), "Only DagRuns are allowed to be subsequent!"
            kwargs["subsequent_field"] = "end_date"
            kwargs["primary_key"] = ["dag_id", "dag_run_id"]
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        if self.subsequent_field and self.test_if_target_table_exists():
            data_from = self.get_max_value_of_column(self.subsequent_field)
            self.log.info("Loading data from {0}".format(data_from))
            if isinstance(data_from, str):
                data_from = parse(data_from)
        else:
            data_from = None
        for batch in self.source_hook.get_data_in_batches(
            endpoint=self.endpoint,
            page_size=self.request_page_size,
            data_from=data_from,
        ):
            self.upload_data(batch)



"""
Airbyte's implementation of this connector has been an important source of information
in the process of building this EWAH connector. Check out Airbyte for more info!
"""

from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC

from datetime import datetime, date, timedelta
from dateutil.parser import parse as parse_datetime
from io import StringIO
import xml.etree.ElementTree as ET
import urllib.parse
import pendulum
import hashlib
import hmac
import boto3
import requests
import time
import pytz
import copy
import gzip
import json
import csv
import random


class EWAHAmazonSellerCentralHook(EWAHBaseHook):
    """
    Implements the Amazon Seller Portal API (SP-API).

    Currently, only the Reporting API is implemented.

    Since the report types vary vastly in implementation, each report type
    has its own method to properly load it. They are supported by common
    methods for authentication and getting the raw report contents.
    """

    # Allowed reports with the alias and various settings
    _REPORT_METADATA = {
        "orders": {
            "report_type": "GET_XML_ALL_ORDERS_DATA_BY_LAST_UPDATE_GENERAL",
            "report_options": {},
            "method_name": "get_order_data_from_reporting_api",
            "primary_key": ["AmazonOrderID"],
            "subsequent_field": "LastUpdatedDate",
            "accepted_strategies": [EC.ES_INCREMENTAL, EC.ES_SUBSEQUENT],
        },
        "sales_and_traffic": {
            "report_type": "GET_SALES_AND_TRAFFIC_REPORT",
            "report_options": {
                "dateGranularity": ["DAY", "WEEK", "MONTH"],
                "asinGranularity": ["PARENT", "CHILD", "SKU"],
            },
            "method_name": "get_sales_and_traffic_data",
            "primary_key": [
                "parentAsin",
                "childAsin",
                "date",
            ],
            "subsequent_field": "date",
            "accepted_strategies": [EC.ES_INCREMENTAL, EC.ES_SUBSEQUENT],
        },
        "fba_returns": {
            # This report does not have a primary key, hence cannot use incremental
            # loading for it - must use full refresh every time
            "report_type": "GET_FBA_FULFILLMENT_CUSTOMER_RETURNS_DATA",
            "report_options": {},
            "method_name": "get_fba_returns_data",
            "primary_key": None,
            "subsequent_field": None,
            "accepted_strategies": [EC.ES_FULL_REFRESH],
        },
        "listings": {  # Full-refresh only
            "report_type": "GET_MERCHANT_LISTINGS_ALL_DATA",
            "report_options": {},
            "method_name": "get_listings",
            "primary_key": None,
            "subsequent_field": None,
            "accepted_strategies": [EC.ES_FULL_REFRESH],
            "ewah_options": {
                "add_bsr": [True, False],
            },
        },
        "inventory": {  # Full-refresh only
            "report_type": "GET_FBA_MYI_UNSUPPRESSED_INVENTORY_DATA",
            "report_options": {},
            "method_name": "get_inventory",
            "primary_key": None,
            "subsequent_field": None,
            "accepted_strategies": [EC.ES_FULL_REFRESH],
        },
        "fee_preview": {  # Full-refresh only
            "report_type": "GET_FBA_ESTIMATED_FBA_FEES_TXT_DATA",
            "report_options": {},
            "method_name": "get_fee_preview",
            "primary_key": None,
            "subsequent_field": None,
            "accepted_strategies": [EC.ES_FULL_REFRESH],
        },
        "settlement_report_v2_financial_transactions": {
            "report_type": "GET_V2_SETTLEMENT_REPORT_DATA_FLAT_FILE_V2",
            "report_options": {},
            "method_name": "get_settlement_report_v2_data",
            "primary_key": [
                "settlement-id",
                "transaction-type",
                "order-id",
                "adjustment-id",
            ],
            "subsequent_field": "report_created_date",
            "accepted_strategies": [EC.ES_SUBSEQUENT],
        },
        "brand_analytics_search_catalog_performance": {
            "report_type": "GET_BRAND_ANALYTICS_SEARCH_CATALOG_PERFORMANCE_REPORT",
            "report_options": {"reportPeriod": ["WEEK"]},
            "method_name": "get_brand_analytics_search_catalog_performance",
            "primary_key": ["asin", "startDate", "endDate"],
            "subsequent_field": "endDate",
            "accepted_strategies": [EC.ES_SUBSEQUENT],
        },
    }

    _ATTR_RELABEL = {}

    conn_name_attr = "ewah_amazon_seller_central_conn_id"
    default_conn_name = "ewah_amazon_seller_central_default"
    conn_type = "ewah_amazon_seller_central"
    hook_name = "EWAH Amazon Seller Central Connection"

    @staticmethod
    def get_ui_field_behaviour():
        # Hide all and use custom fields only
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login", "password"],
            "relabeling": {},
        }

    @staticmethod
    def get_connection_form_widgets():
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import (
            BS3TextFieldWidget,
            BS3PasswordFieldWidget,
        )
        from wtforms import StringField

        return {
            "extra__ewah_amazon_seller_central__aws_access_key_id": StringField(
                "AWS Access Key ID (Optional)", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__aws_secret_access_key": StringField(
                "AWS Secret Access Key (Optional)", widget=BS3PasswordFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__aws_arn_role": StringField(
                "AWS ARN - Role (Optional)", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__lwa_client_id": StringField(
                "LoginWithAmazon (LWA) Client ID", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__lwa_client_secret": StringField(
                "LoginWithAmazon (LWA) Client Secret", widget=BS3PasswordFieldWidget()
            ),
            "extra__ewah_amazon_seller_central__refresh_token": StringField(
                "Refresh Token", widget=BS3PasswordFieldWidget()
            ),
        }

    @staticmethod
    def get_marketplace_details_tuple(marketplace_region):
        base_url = "https://sellingpartnerapi"
        return {
            "AE": (f"{base_url}-eu.amazon.com", "A2VIGQ35RCS4UG", "eu-west-1"),
            "DE": (f"{base_url}-eu.amazon.com", "A1PA6795UKMFR9", "eu-west-1"),
            "PL": (f"{base_url}-eu.amazon.com", "A1C3SOZRARQ6R3", "eu-west-1"),
            "EG": (f"{base_url}-eu.amazon.com", "ARBP9OOSHTCHU", "eu-west-1"),
            "ES": (f"{base_url}-eu.amazon.com", "A1RKKUPIHCS9HS", "eu-west-1"),
            "FR": (f"{base_url}-eu.amazon.com", "A13V1IB3VIYZZH", "eu-west-1"),
            "IN": (f"{base_url}-eu.amazon.com", "A21TJRUUN4KGV", "eu-west-1"),
            "IT": (f"{base_url}-eu.amazon.com", "APJ6JRA9NG5V4", "eu-west-1"),
            "NL": (f"{base_url}-eu.amazon.com", "A1805IZSGTT6HS", "eu-west-1"),
            "SA": (f"{base_url}-eu.amazon.com", "A17E79C6D8DWNP", "eu-west-1"),
            "SE": (f"{base_url}-eu.amazon.com", "A2NODRKZP88ZB9", "eu-west-1"),
            "TR": (f"{base_url}-eu.amazon.com", "A33AVAJ2PDY3EV", "eu-west-1"),
            "UK": (f"{base_url}-eu.amazon.com", "A1F83G8C2ARO7P", "eu-west-1"),
            "AU": (f"{base_url}-fe.amazon.com", "A39IBJ37TRP1C6", "us-west-2"),
            "JP": (f"{base_url}-fe.amazon.com", "A1VC38T7YXB528", "us-west-2"),
            "SG": (f"{base_url}-fe.amazon.com", "A19VAU5U5O7RUS", "us-west-2"),
            "US": (f"{base_url}-na.amazon.com", "ATVPDKIKX0DER", "us-east-1"),
            "BR": (f"{base_url}-na.amazon.com", "A2Q3Y263D00KWC", "us-east-1"),
            "CA": (f"{base_url}-na.amazon.com", "A2EUQ1WTGCTBG2", "us-east-1"),
            "MX": (f"{base_url}-na.amazon.com", "A1AM78C64UM0Y8", "us-east-1"),
            "GB": (f"{base_url}-eu.amazon.com", "A1F83G8C2ARO7P", "eu-west-1"),
        }[marketplace_region]

    @classmethod
    def get_cleaner_callables(cls):
        def string_to_date(row):
            for key, value in row.items():
                if key.lower().endswith("date") and not isinstance(
                    value, (date, datetime)
                ):
                    # Cast all dates
                    if value:  # Don't parse NoneType or empty string
                        row[key] = parse_datetime(value)
            if row.get("parentAsin") and not row.get("childAsin"):
                row["childAsin"] = "n.a."
            return row

        return string_to_date

    @classmethod
    def validate_marketplace_region(
        cls, marketplace_region, raise_on_failure=False, allow_lists=False
    ):
        # Validate whether a region exists. Return boolean or raise exception.

        # Optionally check all regions in a list of regions
        if allow_lists and isinstance(marketplace_region, list):
            for region in marketplace_region:
                if not cls.validate_marketplace_region(
                    region, raise_on_failure=raise_on_failure
                ):
                    return False
            # If execution gets here, it validated all regions of the list successfully
            return True

        # Wrong data type, not valid by definition
        elif not isinstance(marketplace_region, str):
            return False

        # Now validate!
        try:
            cls.get_marketplace_details_tuple(marketplace_region)
        except:
            if raise_on_failure:
                raise
            else:
                return False
        finally:
            return True

    @staticmethod
    def _sign_msg(key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    @property
    def access_token(self):
        if (
            not hasattr(self, "_access_token")
            or pendulum.now() > self._access_token_expires_at
        ):
            self.log.info("Requesting new access token...")
            requested_at = pendulum.now()
            url = "https://api.amazon.com/auth/o2/token"
            payload = {
                "grant_type": "refresh_token",
                "client_id": self.conn.lwa_client_id,
                "client_secret": self.conn.lwa_client_secret,
                "refresh_token": self.conn.refresh_token,
            }
            response = requests.post(url, data=payload)
            assert response.status_code == 200, response.text
            response_data = response.json()
            self._access_token = response_data["access_token"]
            self._access_token_expires_at = requested_at.add(
                # add a 15% error margin to avoid expiration during a call
                seconds=(response_data["expires_in"] * 0.85),
            )
            self.log.info(
                "New Amazon oauth access token is valid until {0}...".format(
                    self._access_token_expires_at.isoformat()
                )
            )
        return self._access_token

    @property
    def use_legacy_authentication(self):
        legacy_secrets = [
            self.conn.aws_access_key_id,
            self.conn.aws_secret_access_key,
            self.conn.aws_arn_role,
        ]
        if any([secret in (None, "") for secret in legacy_secrets]):
            if all([secret in (None, "") for secret in legacy_secrets]):
                return False
            raise Exception(
                "Some, but not all, legacy secrets are used! Either use none at all or fill all out (AWS access key, secret key, arn role)"
            )
        return True

    @property
    def boto3_role_credentials(self):
        if (
            not hasattr(self, "_boto3_role_credentials")
            or pendulum.now() > self._boto3_session_expires_at
        ):
            self.log.info("Logging in to AWS...")
            boto3_client = boto3.client(
                "sts",
                aws_access_key_id=self.conn.aws_access_key_id,
                aws_secret_access_key=self.conn.aws_secret_access_key,
            )
            role = boto3_client.assume_role(
                RoleArn=self.conn.aws_arn_role, RoleSessionName="guid"
            )
            self._boto3_role_credentials = role["Credentials"]
            # Add a 15% error margin to avoid expiration during a call
            self._boto3_session_expires_at = pendulum.now() + 0.85 * (
                role["Credentials"]["Expiration"] - pendulum.now()
            )
            self.log.info(
                "Logged in to AWS! Login expires at: {0}".format(
                    self._boto3_session_expires_at.isoformat()
                )
            )
        return self._boto3_role_credentials

    @property
    def aws_session_token(self):
        return self.boto3_role_credentials.get("SessionToken")

    @property
    def aws_session_secret_access_key(self):
        return self.boto3_role_credentials.get("SecretAccessKey")

    @property
    def aws_session_access_key_id(self):
        return self.boto3_role_credentials.get("AccessKeyId")

    def generate_request_headers(self, url, method, region, params=None, body=None):
        # Returns the headers dict to authorize a request for the SP-API
        if not self.use_legacy_authentication:
            return {
                "x-amz-access-token": self.access_token,
                "Content-Type": "application/json",
            }

        # Below is the legacy authentication header generation

        assert method in ["GET", "POST"]
        params = params or {}

        current_ts = pendulum.now("utc")
        amz_date = current_ts.strftime("%Y%m%dT%H%M%SZ")
        datestamp = current_ts.strftime("%Y%m%d")
        url_parsed = urllib.parse.urlparse(url)
        uri = urllib.parse.quote(url_parsed.path)
        host = url_parsed.hostname
        service = "execute-api"

        canonical_querystring = "&".join(
            map(
                lambda param: "=".join(param),
                sorted(
                    [
                        [str(key), urllib.parse.quote_plus(str(value))]
                        for key, value in params.items()
                    ]
                ),
            )
        )

        headers_to_sign = {
            "host": host,
            "x-amz-date": amz_date,
            "x-amz-security-token": self.aws_session_token,
        }
        ordered_headers = dict(sorted(headers_to_sign.items(), key=lambda h: h[0]))
        canonical_headers = "".join(
            map(lambda h: ":".join(h) + "\n", ordered_headers.items())
        )
        signed_headers = ";".join(ordered_headers.keys())

        if body:
            payload_hash = hashlib.sha256(
                json.dumps(dict(sorted(body.items(), key=lambda h: h[0]))).encode(
                    "utf-8"
                )
            ).hexdigest()
        else:
            payload_hash = hashlib.sha256("".encode("utf-8")).hexdigest()

        canonical_request = "\n".join(
            [
                method,
                uri,
                canonical_querystring,
                canonical_headers,
                signed_headers,
                payload_hash,
            ]
        )

        credential_scope = "/".join([datestamp, region, service, "aws4_request"])
        string_to_sign = "\n".join(
            [
                "AWS4-HMAC-SHA256",
                amz_date,
                credential_scope,
                hashlib.sha256(canonical_request.encode("utf-8")).hexdigest(),
            ]
        )

        datestamp_signed = self._sign_msg(
            ("AWS4" + self.aws_session_secret_access_key).encode("utf-8"), datestamp
        )
        region_signed = self._sign_msg(datestamp_signed, region)
        service_signed = self._sign_msg(region_signed, service)
        aws4_request_signed = self._sign_msg(service_signed, "aws4_request")
        signature = hmac.new(
            aws4_request_signed, string_to_sign.encode("utf-8"), hashlib.sha256
        ).hexdigest()
        authorization_header = (
            "AWS4-HMAC-SHA256 Credential={}/{}, SignedHeaders={}, Signature={}".format(
                self.aws_session_access_key_id,
                credential_scope,
                signed_headers,
                signature,
            )
        )

        return {
            "content-type": "application/json",
            "host": host,
            "user-agent": "python-requests",
            "x-amz-access-token": self.access_token,
            "x-amz-date": amz_date,
            "x-amz-security-token": self.aws_session_token,
            "authorization": authorization_header,
        }

    def get_report_data(
        self,
        marketplace_region,
        report_name,
        data_from=None,
        data_until=None,
        report_options=None,
    ):
        # This method calls the reporting API to fetch data on a report
        # Use this method when writing the individual reports' fetching methods
        # Returns the report content unaltered (only decompressed, if applicable)

        report_metadata = self._REPORT_METADATA[report_name]
        report_type = report_metadata["report_type"]

        endpoint, marketplace_id, region = self.get_marketplace_details_tuple(
            marketplace_region
        )
        url = endpoint + "/reports/2021-06-30/reports"
        body = {
            "marketplaceIds": [marketplace_id],
            "reportType": report_type,
        }

        if report_options:
            available_options = report_metadata.get("report_options", {})
            assert isinstance(report_options, dict), "report_options must be a dic!"
            for option, value in report_options.items():
                assert (
                    option in available_options.keys()
                ), f"Invalid report option '{option}'!"
                assert (
                    value in available_options[option]
                ), f"Invalid value '{value}' for option '{option}'!"
            body["reportOptions"] = report_options

        if data_from:
            if isinstance(data_from, (date, datetime)):
                data_from = data_from.isoformat()
            body["dataStartTime"] = data_from

        if data_until:
            if isinstance(data_until, (date, datetime)):
                data_until = data_until.isoformat()
            body["dataEndTime"] = data_until

        # The order of the items in the body is relevant for proper authorization
        body = dict(sorted(body.items(), key=lambda h: h[0]))
        self.log.info("Creating the report via POST request now... " f"Payload: {body}")
        response = requests.post(
            url,
            json=body,
            headers=self.generate_request_headers(
                url=url, method="POST", region=region, body=body
            ),
        )
        assert response.status_code == 202, response.text
        report_id = response.json()["reportId"]

        # Fetch the report Part 1: Get the document ID
        self.log.info(f"Report created, ID: {report_id}.")
        url = endpoint + "/reports/2021-06-30/reports/" + report_id
        tries = 0
        sleep_base = 5
        sleep_for = sleep_base
        done = False
        time.sleep(sleep_for)
        while tries < 10:
            tries += 1
            # Exponential increase of the time we are waiting
            sleep_for = sleep_base * (2**tries)
            self.log.info(f"Trying ({tries}/10) to fetch the document id...")
            response = requests.get(
                url,
                headers=self.generate_request_headers(
                    url=url, method="GET", region=region
                ),
            )
            assert response.status_code == 200, response.text
            status = response.json().get("processingStatus")
            if status == "DONE":
                done = True
                break
            elif status in ("IN_PROGRESS", "IN_QUEUE"):
                # Wait a bit, try again
                pass
            elif status == "CANCELLED":
                # There is no data to report on
                return
            elif status == "FATAL":
                raise Exception("FATAL ERROR! Request response: " + response.text)
            else:
                raise Exception(
                    "Unexpected processing status! Request response: " + response.text
                )
            self.log.info(
                f"Unable to fetch document, status: {status}."
                f" Waiting for {sleep_for}s..."
            )
            time.sleep(sleep_for)

        assert response.json().get("processingStatus") == "DONE", response.text
        report_document_id = response.json()["reportDocumentId"]

        # Fetch the report Part 2: Get the document url
        self.log.info(
            f"Got document ID: {report_document_id}." " Fetching document url now."
        )
        url = endpoint + "/reports/2021-06-30/documents/" + report_document_id
        response = requests.get(
            url,
            headers=self.generate_request_headers(url=url, method="GET", region=region),
        )
        assert response.status_code == 200, response.text
        response_data = response.json()
        assert (
            response_data.get("compressionAlgorithm") is None
            or response_data["compressionAlgorithm"] == "GZIP"
        ), "Invalid Compression Algorithm: {0}".format(
            response_data.get("compressionAlgorithm")
        )
        is_gzip = response_data.get("compressionAlgorithm") == "GZIP"
        document_url = response_data["url"]

        # Fetch the report Part 3: Download the document
        self.log.info(f"Got document URL. Now downloading document.")
        document_response = requests.get(document_url, allow_redirects=True)
        # Fix check (might be temporarily):
        # The compressionAlgorithm parameter is not reliable at the moment.
        # Its set, but the content is not always zipped, which we check here.
        is_zipped = document_response.content.startswith(b"\x1f\x8b")

        assert document_response.status_code == 200, document_response.text
        if is_zipped:
            document_content = gzip.decompress(document_response.content)
        else:
            document_content = document_response.content

        return document_content

    def get_order_data_from_reporting_api(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        if (data_until - data_from) > timedelta(days=29):
            # Special case: Order reports can't be fetched
            # for a time period greater than 30 days.
            # Loop over the entire period in steps of 29 days.
            while data_from < data_until:
                for batch in self.get_order_data_from_reporting_api(
                    marketplace_region=marketplace_region,
                    report_name=report_name,
                    data_from=data_from,
                    data_until=min(data_until, data_from + timedelta(days=29)),
                    report_options=report_options,
                    batch_size=batch_size,
                ):
                    yield batch
                data_from += timedelta(days=29)
            # we're done here after looping over the period
            return

        self.log.info(
            "Fetching order data from {0} to {1}...".format(
                data_from.isoformat(), data_until.isoformat()
            )
        )

        def simple_xml_to_json(xml, depth=1):
            # Takes xml and turns it to a (nested) dictionary
            response = {}
            scalars = [
                "AmazonOrderID",
                "MerchantOrderID",
                "OrderStatus",
                "SalesChannel",
                "IsBusinessOrder",
                "IsIba",
                "LastUpdatedDate",
                "PurchaseDate",
            ]
            for child in list(xml):
                if not response.get(child.tag):
                    # Everything becomes a list in order to be consistent,
                    # even if length is always 1 for some children
                    response[child.tag] = []
                if len(list(child)) > 0:
                    # child is also an object
                    response[child.tag].append(simple_xml_to_json(child, depth + 1))
                else:
                    # tag is a scalar, add it
                    if (  # depth == 1 is Message, 2 is Order
                        depth == 3 and child.tag in scalars
                    ):
                        # Special cases - these are never lists
                        response[child.tag] = child.text
                    else:
                        response[child.tag].append(child.text)
            return response

        # Note: get_report_data may return None if there is no new data!
        data_string = (
            self.get_report_data(
                marketplace_region,
                report_name,
                data_from,
                data_until,
                report_options,
            )
            or b""
        ).decode()
        if data_string:
            self.log.info("Turning response XML into JSON...")
            raw_data = simple_xml_to_json(ET.fromstring(data_string)).get("Message", [])
        else:
            # No data to provide
            raw_data = []

        # Respect batch size kwarg and prettify data structure
        data = []
        i = 0
        while raw_data:
            i += 1
            # Due to the conversion logic, Message is a Dict of which we only want
            # Order, and Order is a list that is always of size 1.
            # We only want a list of Order as final result, though.
            data.append(raw_data.pop(0)["Order"][0])
            if i == batch_size:
                yield data
                data = []
                i = 0
        if data:
            yield data

    def get_sales_and_traffic_data(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,  # is ignored in this specific function
    ):
        delta_day = timedelta(days=1)

        # This report fetches data in full day periods
        if isinstance(data_from, datetime):
            data_from = data_from.date()
        if isinstance(data_until, datetime):
            data_until = data_until.date()

        while True:
            # Sales and traffic report needs to be requested individually per day
            started = datetime.now()  # used in the next iteration, if applicable
            data_from_dt = data_from
            if isinstance(data_from_dt, pendulum.Date):
                # Uploader class doesn't like Pendulum (data_from_dt is added to data)
                data_from_dt = date(
                    data_from_dt.year,
                    data_from_dt.month,
                    data_from_dt.day,
                )

            # Note that if no data is returned (e.g., timeframe too early), the get_report_data method returns None
            data_raw = (
                self.get_report_data(
                    marketplace_region,
                    report_name,
                    data_from,
                    data_from,
                    report_options,
                )
                or b""
            ).decode()
            data = json.loads(data_raw or "{}").get("salesAndTrafficByAsin", [])
            for datum in data:
                # add the requested day to all rows
                datum["date"] = data_from_dt
            yield data

            data_from += delta_day
            if data_from > data_until:
                break

            self.log.info("Delaying execution for up to a minute...")
            # Delaying to avoid hitting API request rate limits
            # --> Wait 1 minute between iterations
            time.sleep(
                max(
                    0,
                    (started + timedelta(seconds=60) - datetime.now()).total_seconds(),
                )
            )

    def get_fba_returns_data(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        self.log.info(
            "Fetching FBA returns data from {0} to {1}...".format(
                data_from.isoformat(), data_until.isoformat()
            )
        )

        # Note: This report is only available as full refresh, however
        # it still requires data_from (ideally, static) and data_until
        # (ideally, current timestamp).
        assert data_from, "Requires a minimum date parameter!"
        # data_until is defaulted to current timestamp

        # Return data from CSV in batches
        data_io = StringIO(
            self.get_report_data(
                marketplace_region,
                report_name,
                data_from,
                data_until or datetime.utcnow().replace(tzinfo=pytz.utc),
                report_options,
            ).decode()
        )  # TODO: check if latin-1?
        csv_reader = csv.DictReader(data_io, delimiter="\t")
        data = []
        i = 0
        for row in csv_reader:
            i += 1
            data.append(row)
            if i == batch_size:
                yield data
                i = 0
                data = []
        if data:
            yield data

    def get_listings(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        self.log.info("Fetching Listings. Ignoring datetimes if any.")
        data_io = StringIO(
            self.get_report_data(
                marketplace_region,
                report_name,
                None,
                None,
                report_options,
            ).decode("latin-1")
        )
        csv_reader = csv.DictReader(data_io, delimiter="\t")
        data = []
        i = 0
        for row in csv_reader:
            i += 1
            if ewah_options and ewah_options.get("add_bsr"):
                # Make a request to add the BSR at this point
                asin = row.get("asin1")
                if asin:
                    self.log.info("Fetching additional catalogue data for a listing...")
                    row.update(self.get_listing_details(marketplace_region, asin) or {})
            data.append(row)
            if i == batch_size:
                yield data
                i = 0
                data = []
        if data:
            yield data

    def get_inventory(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        self.log.info("Fetching Inventory. Ignoring datetimes if any.")
        data_io = StringIO(
            self.get_report_data(
                marketplace_region,
                report_name,
                None,
                None,
                report_options,
            ).decode("latin-1")
        )
        csv_reader = csv.DictReader(data_io, delimiter="\t")
        data = []
        i = 0
        for row in csv_reader:
            i += 1
            data.append(row)
            if i == batch_size:
                yield data
                i = 0
                data = []
        if data:
            yield data

    def get_fee_preview(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        self.log.info("Fetching Fee Preview Report...")

        # The FBA Fee Preview Report (GET_FBA_ESTIMATED_FBA_FEES_TXT_DATA)
        # "contains the estimated Amazon Selling and Fulfillment Fees for the seller's FBA
        # inventory with active offers. The content is updated at least once every
        # 72 hours. To successfully generate a report, specify the StartDate parameter
        # for a minimum 72 hours prior to NOW and EndDate to NOW."
        # https://developer-docs.amazon.com/sp-api/docs/report-type-values-fba#fba-payments-reports

        if not data_from:
            data_from = datetime.utcnow().replace(tzinfo=pytz.utc) - timedelta(hours=72)
            self.log.info(
                f"No data_from provided, setting to 72 hours ago: {data_from}"
            )
        else:
            self.log.info(f"Using provided data_from: {data_from}")

        document_content = self.get_report_data(
            marketplace_region,
            report_name,
            data_from,
            data_until,
            report_options,
        )

        if document_content is None:
            self.log.warning("No document content returned")
            return

        data_io = StringIO(document_content.decode("latin-1"))
        csv_reader = csv.DictReader(data_io, delimiter="\t")
        data = []
        i = 0
        for row in csv_reader:
            i += 1
            data.append(row)
            if i == batch_size:
                yield data
                i = 0
                data = []
        if data:
            yield data

    def get_brand_analytics_search_catalog_performance(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        """Fetch Brand Analytics Search Catalog Performance Report from Amazon Seller Central.

        This report provides insights into search catalog performance for brand analytics.
        Only WEEK reportPeriod is supported.

        Requirements:
        - Requests must include reportPeriod=WEEK in the reportsOptions
        - dataStartTime must be a Sunday and dataEndTime must be a Saturday
        - If the first date (data_from) is not a Sunday, move it forward to the next available Sunday.
        """
        self.log.info("Fetching Brand Analytics Search Catalog Performance Report...")

        # Ensure report_options is a dictionary
        if report_options is None:
            report_options = {}

        # Validate that reportPeriod is provided
        if "reportPeriod" not in report_options:
            raise ValueError(
                "reportPeriod must be specified in report_options for Brand Analytics Search Catalog Performance Report"
            )

        report_period = report_options["reportPeriod"]

        # Validate reportPeriod
        if report_period != "WEEK":
            raise ValueError(
                f"Only reportPeriod=WEEK is supported. Got: {report_period}"
            )

        # Normalize data_from to date object if it's a datetime
        if isinstance(data_from, datetime):
            data_from = data_from.date()

        # Handle date_from: Weekly reports need to start on a Sunday,
        # so adjust to next Sunday if not already a Sunday
        original_date = data_from
        if data_from.weekday() != 6:  # Not Sunday
            # Calculate days until next Sunday
            # weekday: 0=Mon, 1=Tue, 2=Wed, 3=Thu, 4=Fri, 5=Sat, 6=Sun
            days_until_sunday = 6 - data_from.weekday()
            data_from = data_from + timedelta(days=days_until_sunday)
            self.log.info(
                f"Adjusted data_from from {original_date.strftime('%Y-%m-%d')} "
                f"({original_date.strftime('%A')}) to next Sunday: {data_from.strftime('%Y-%m-%d')}"
            )

        # Set up weekly data fetching
        start_date = data_from

        # Calculate the last available Saturday
        today = date.today()
        if today.weekday() == 5:  # Today is Saturday
            last_saturday = today
        else:  # Find the most recent Saturday
            days_since_saturday = (
                today.weekday() + 2
            ) % 7  # Calculate days since last Saturday
            last_saturday = today - timedelta(days=days_since_saturday)

        self.log.info(
            f"Fetching weekly data from {start_date.strftime('%Y-%m-%d')} to {last_saturday.strftime('%Y-%m-%d')}"
        )

        # Process each week from start_date to last_saturday
        current_sunday = start_date
        while current_sunday <= last_saturday:
            current_saturday = current_sunday + timedelta(days=6)

            # End if this week would go beyond our end date
            if current_saturday > last_saturday:
                break

            self.log.info(
                f"Processing week: {current_sunday.strftime('%Y-%m-%d')} to {current_saturday.strftime('%Y-%m-%d')}"
            )

            # Process the report data as JSON for this week
            max_retries = 5
            base_delay = 30
            data_raw = None

            try:
                for attempt in range(max_retries):
                    try:
                        data_raw = (
                            self.get_report_data(
                                marketplace_region,
                                report_name,
                                current_sunday,
                                current_saturday,
                                report_options,
                            )
                            or b""
                        ).decode()
                        # Success - break out of retry loop
                        break
                    except (AssertionError, Exception) as e:
                        error_text = str(e)
                        # Check if it's a QuotaExceeded error (can occur at any stage of get_report_data)
                        is_quota_error = (
                            "QuotaExceeded" in error_text
                            or "quota" in error_text.lower()
                        )

                        if is_quota_error and attempt < max_retries - 1:
                            # Exponential backoff: delay = base_delay * (2^attempt)
                            delay = base_delay * (2**attempt)
                            self.log.warning(
                                f"Quota exceeded for week {current_sunday.strftime('%Y-%m-%d')} to {current_saturday.strftime('%Y-%m-%d')}. "
                                f"Attempt {attempt + 1}/{max_retries}. Waiting {delay} seconds before retry..."
                            )
                            time.sleep(delay)
                            continue
                        # If max retries exceeded or non-quota error, raise
                        raise
            except Exception as e:
                # We assume that if the error is FATAL, it is because the report is not available yet
                if "FATAL" in str(e):
                    self.log.error(
                        f"Fatal error encountered for week {current_sunday.strftime('%Y-%m-%d')} to {current_saturday.strftime('%Y-%m-%d')}: {e}\n Skipping week..."
                    )
                    current_sunday += timedelta(days=7)
                    continue
                else:
                    raise e

            if not data_raw:
                self.log.info(
                    f"No data returned for week {current_sunday.strftime('%Y-%m-%d')} to {current_saturday.strftime('%Y-%m-%d')}"
                )
                # Move to next week
                current_sunday += timedelta(days=7)
                continue

            json_data = json.loads(data_raw)
            raw_data = json_data.get("dataByAsin", [])

            # Process and flatten the data for this week
            data = []
            i = 0
            for item in raw_data:
                # Extract top-level fields
                flattened_row = {
                    "startDate": item.get("startDate"),
                    "endDate": item.get("endDate"),
                    "asin": item.get("asin"),
                }

                # Flatten impressionData fields
                impression_data = item.get("impressionData", {})
                if impression_data:
                    flattened_row.update(
                        {
                            "impressionCount": impression_data.get("impressionCount"),
                            "impressionMedianPrice": impression_data.get(
                                "impressionMedianPrice"
                            ),
                            "sameDayShippingImpressionCount": impression_data.get(
                                "sameDayShippingImpressionCount"
                            ),
                            "oneDayShippingImpressionCount": impression_data.get(
                                "oneDayShippingImpressionCount"
                            ),
                            "twoDayShippingImpressionCount": impression_data.get(
                                "twoDayShippingImpressionCount"
                            ),
                        }
                    )

                # Flatten clickData fields
                click_data = item.get("clickData", {})
                if click_data:
                    flattened_row.update(
                        {
                            "clickCount": click_data.get("clickCount"),
                            "clickRate": click_data.get("clickRate"),
                            "clickedMedianPrice": click_data.get("clickedMedianPrice"),
                            "sameDayShippingClickCount": click_data.get(
                                "sameDayShippingClickCount"
                            ),
                            "oneDayShippingClickCount": click_data.get(
                                "oneDayShippingClickCount"
                            ),
                            "twoDayShippingClickCount": click_data.get(
                                "twoDayShippingClickCount"
                            ),
                        }
                    )

                # Flatten cartAddData fields
                cart_add_data = item.get("cartAddData", {})
                if cart_add_data:
                    flattened_row.update(
                        {
                            "cartAddCount": cart_add_data.get("cartAddCount"),
                            "cartAddedMedianPrice": cart_add_data.get(
                                "cartAddedMedianPrice"
                            ),
                            "sameDayShippingCartAddCount": cart_add_data.get(
                                "sameDayShippingCartAddCount"
                            ),
                            "oneDayShippingCartAddCount": cart_add_data.get(
                                "oneDayShippingCartAddCount"
                            ),
                            "twoDayShippingCartAddCount": cart_add_data.get(
                                "twoDayShippingCartAddCount"
                            ),
                        }
                    )

                # Flatten purchaseData fields
                purchase_data = item.get("purchaseData", {})
                if purchase_data:
                    flattened_row.update(
                        {
                            "purchaseCount": purchase_data.get("purchaseCount"),
                            "searchTrafficSales": purchase_data.get(
                                "searchTrafficSales"
                            ),
                            "conversionRate": purchase_data.get("conversionRate"),
                            "purchaseMedianPrice": purchase_data.get(
                                "purchaseMedianPrice"
                            ),
                            "sameDayShippingPurchaseCount": purchase_data.get(
                                "sameDayShippingPurchaseCount"
                            ),
                            "oneDayShippingPurchaseCount": purchase_data.get(
                                "oneDayShippingPurchaseCount"
                            ),
                            "twoDayShippingPurchaseCount": purchase_data.get(
                                "twoDayShippingPurchaseCount"
                            ),
                        }
                    )

                data.append(flattened_row)
                i += 1

                if i == batch_size:
                    yield data
                    i = 0
                    data = []

            if data:
                yield data

            # Log completion before moving to next week
            self.log.info(
                f"Retrieved report for week {current_sunday.strftime('%Y-%m-%d')} to {current_saturday.strftime('%Y-%m-%d')}"
            )

            # Move to next week
            current_sunday += timedelta(days=7)

            # Add a small delay between weekly requests to respect API rate limits
            if current_sunday <= last_saturday:
                self.log.info("Waiting 2 seconds before processing next week...")
                time.sleep(2)

    def get_settlement_report_v2_data(
        self,
        marketplace_region,
        report_name,
        data_from,
        data_until,
        report_options=None,
        ewah_options=None,
        batch_size=20000,
    ):
        """Fetch settlement report v2 data from Amazon Seller Central.

        This report contains detailed financial transaction data including:
        - Settlement ID
        - Transaction type
        - Order ID
        - Posted date
        - Other financial details (specifically, all fee types)

        ! Note:
        Unlike the other report types in this hook, the settlement report works differently:
        1. Settlement reports are automatically generated by Amazon on a regular basis
        2. Instead of requesting a new report to be generated, we search for and download
           existing reports that were already created by Amazon
        3. The date range parameters (data_from and data_until) are used to find reports
           that were created within that period, not to generate new reports
        4. Multiple reports may be returned if they exist in the specified date range
        5. Each report contains a complete settlement period's worth of data
        6. There is a 90-day limit on how far back you can retrieve reports
        7. The first row of the report is a summary row that has no transaction details
        and has to be skipped in order to have a consistent data structure that allows for
        incremental loading via the "posted-date" field (which is NULL for summary rows
        -> cannot be parsed as datetime object).

        This is why this endpoint uses a different approach than other report types:
        - It uses GET instead of POST to search for existing reports
        - It processes multiple reports if they exist in the date range
        - It doesn't require report generation, just report retrieval
        """
        self.log.info(
            "Fetching Settlement Report V2 data from {0} to {1}...".format(
                data_from.isoformat() if data_from else "beginning",
                data_until.isoformat() if data_until else "now",
            )
        )

        # Settlement reports require a data_from parameter
        assert data_from, "Requires a minimum date parameter!"

        # Ensure data_until is not None
        if not data_until:
            data_until = datetime.utcnow().replace(tzinfo=pytz.utc)
            self.log.info(f"No data_until provided, using current time: {data_until}")

        # Adjust data_from if it's more than 90 days old
        min_possible_date = datetime.utcnow().replace(tzinfo=pytz.utc) - timedelta(
            days=90
        )
        if data_from < min_possible_date:
            self.log.info(
                f"Requested start date {data_from} is more than 90 days old. "
                f"Adjusting to {min_possible_date} to comply with API limitations."
            )
            data_from = min_possible_date

        endpoint, marketplace_id, region = self.get_marketplace_details_tuple(
            marketplace_region
        )

        # Step 1: Get access token
        access_token = self.access_token

        # Step 2: Search for available reports
        url = f"{endpoint}/reports/2021-06-30/reports"
        params = {
            "reportTypes": ["GET_V2_SETTLEMENT_REPORT_DATA_FLAT_FILE_V2"],
            "pageSize": 100,
            "createdSince": data_from.isoformat(),
            "createdUntil": data_until.isoformat(),
        }
        headers = {
            "x-amz-access-token": access_token,
            "Content-Type": "application/json",
        }

        self.log.info(
            f"Searching for settlement reports created between {data_from.isoformat()} "
            f"and {data_until.isoformat()}..."
        )
        max_retries = 3
        delay = 60

        for attempt in range(max_retries):
            try:
                response = requests.get(url, headers=headers, params=params)
                if response.status_code == 200:
                    break
                elif response.status_code == 429 or "QuotaExceeded" in response.text:
                    if attempt < max_retries - 1:
                        self.log.warning(
                            f"[Step 1/4] Quota exceeded. Waiting {delay} seconds before retry..."
                        )
                        time.sleep(delay)
                        continue
                response.raise_for_status()
            except Exception as e:
                if attempt == max_retries - 1:
                    raise
                continue

        assert response.status_code == 200, response.text
        reports = response.json()["reports"]

        if not reports:
            self.log.warning("No settlement reports found in the specified date range")
            return

        self.log.info(f"Found {len(reports)} settlement reports")

        # Process each report
        for report_index, report in enumerate(reports, 1):
            report_id = report["reportId"]
            self.log.info(
                f"Processing report {report_id} ({report_index}/{len(reports)})..."
            )

            # Add delay to respect SP-API rate limit of 0.5 requests per second
            if report_index > 1:
                time.sleep(
                    2
                )  # 2 seconds between requests to stay under 0.5 requests/second

            # Get document ID
            url = f"{endpoint}/reports/2021-06-30/reports/{report_id}"
            for attempt in range(max_retries):
                try:
                    response = requests.get(url, headers=headers)
                    response.raise_for_status()
                    break
                except Exception as e:
                    if attempt == max_retries - 1:
                        raise
                    self.log.warning(
                        f"Report {report_id} - Step 1/3: API request failed. Retrying in {delay} seconds..."
                    )
                    time.sleep(delay)
                    continue

            document_id = response.json()["reportDocumentId"]
            self.log.info(
                f"Report {report_id} - Step 1/3: Got document ID: {document_id}"
            )

            # Get document URL
            url = f"{endpoint}/reports/2021-06-30/documents/{document_id}"
            for attempt in range(max_retries):
                try:
                    response = requests.get(url, headers=headers)
                    response.raise_for_status()
                    break
                except Exception as e:
                    if attempt == max_retries - 1:
                        raise
                    self.log.warning(
                        f"Report {report_id} - Step 2/3: API request failed. Retrying in {delay} seconds..."
                    )
                    time.sleep(delay)
                    continue

            document_url = response.json()["url"]
            self.log.info(f"Report {report_id} - Step 2/3: Got document URL")

            # Download and process the report
            document_response = requests.get(document_url)
            assert document_response.status_code == 200, document_response.text
            self.log.info(
                f"Report {report_id} - Step 3/3: Downloading and processing report data..."
            )

            # Process the CSV data
            data_io = StringIO(document_response.content.decode("latin-1"))
            csv_reader = csv.DictReader(data_io, delimiter="\t")
            data = []
            i = 0
            first_row = True

            # Reset the CSV reader to start from beginning
            data_io.seek(0)
            csv_reader = csv.DictReader(data_io, delimiter="\t")

            # Get and parse report creation date
            report_created_date = datetime.fromisoformat(
                report.get("createdTime").replace("Z", "+00:00")
            )
            self.log.info(f"Report creation date: {report_created_date.isoformat()}")

            for row in csv_reader:
                # Skip summary rows (has settlement-id but no transaction details)
                # as they would break the data structure needed for incremental loading
                if row.get("settlement-id") and not row.get("transaction-type"):
                    continue

                i += 1
                # Add the report creation date to each row -> for incremental loading
                row["report_created_date"] = report_created_date

                if first_row:
                    self.log.info(f"Settlement ID: {row['settlement-id']}")
                    first_row = False

                data.append(row)
                if i == batch_size:
                    self.log.info(f"Yielding batch of {len(data)} rows.")
                    yield data
                    i = 0
                    data = []

            # Yield any remaining transaction data that hasn't been yielded yet due to batch size
            if data:
                yield data

    def get_data_from_reporting_api_in_batches(
        self,
        marketplace_region,
        report_name,
        data_from=None,
        data_until=None,
        report_options=None,
        ewah_options=None,
        batch_size=10000,
    ):
        error_msg = """Invalid report name {1}! Valid options:
        \n\t- {0}
        """.format(
            "\n\t- ".join(self._REPORT_METADATA.keys()),
            report_name,
        )
        assert report_name in self._REPORT_METADATA.keys(), error_msg

        if isinstance(marketplace_region, list):
            # Special case - multiple marketplace_regions!
            # If we have a list of marketplace_regions, call self
            # once per region and add the region to the data.
            for region in marketplace_region:
                for batch in self.get_data_from_reporting_api_in_batches(
                    marketplace_region=region,
                    report_name=report_name,
                    data_from=data_from,
                    data_until=data_until,
                    report_options=report_options,
                    ewah_options=ewah_options,
                    batch_size=batch_size,
                ):
                    for datum in batch:
                        datum["ewah_marketplace_region"] = region
                    yield batch
            return
        elif not isinstance(marketplace_region, str):
            raise Exception("'marketplace_region' must be string or list of strings!")

        self.log.info(
            f"Fetching report '{report_name}' for region '{marketplace_region}' "
            f"between {data_from} and {data_until}.",
        )

        method = getattr(self, self._REPORT_METADATA[report_name]["method_name"])
        data = []
        for batch in method(
            marketplace_region,
            report_name,
            data_from,
            data_until,
            report_options,
            ewah_options,
        ):
            data += batch
            if len(data) >= batch_size:
                yield data
                data = []
        if data:
            # Last batch may otherwise not be yielded if below threshold of batch_size
            yield data

    def get_listing_details(self, marketplace_region, asin):
        # For a single ASIN, retrieve catalogue data, e.g. BSRs
        # Returns None or a dictionarty containing the "includedData"
        url, marketplace_id, region = self.get_marketplace_details_tuple(
            marketplace_region
        )
        url = "".join([url, "/catalog/2022-04-01/items/", asin])
        params = {
            "marketplaceIds": marketplace_id,
            "includedData": ",".join(
                [
                    "salesRanks",
                    "summaries",
                    "relationships",
                    "identifiers",
                    "attributes",
                    "dimensions",
                ]
            ),
        }

        # Backoff strategy parameters
        max_retries = 5
        base_delay = 2
        max_delay = 60
        jitter = 0.1

        for attempt in range(max_retries):
            requested_at = datetime.utcnow()
            response = requests.get(
                url,
                params=params,
                headers=self.generate_request_headers(
                    url=url, method="GET", region=region, params=params
                ),
            )

            if response.status_code == 200:
                # If successful, respect endpoint response rate limit of 2 requests per second
                # Run 1 request per second, just to be sure in case of conflicts
                time.sleep(
                    max(
                        0,
                        (
                            requested_at + timedelta(seconds=1) - datetime.utcnow()
                        ).total_seconds(),
                    )
                )
                return {
                    f"catalogue_{key}": value for key, value in response.json().items()
                }

            elif response.status_code == 429:
                # If rate limit exceeded even though we respected the rate limit
                # We'll retry with exponential backoff
                if attempt < max_retries - 1:
                    # Calculate delay with exponential backoff and jitter
                    delay = min(max_delay, base_delay * (2**attempt))
                    jitter_amount = delay * jitter * (2 * random.random() - 1)
                    total_delay = max(0, delay + jitter_amount)

                    self.log.warning(
                        f"Rate limit exceeded. Attempt {attempt + 1}/{max_retries}. "
                        f"Retrying in {total_delay:.2f} seconds..."
                    )
                    time.sleep(total_delay)
                    continue
            else:
                # Check for NOT_FOUND in the response body
                try:
                    error_data = response.json()
                    if error_data.get("errors", [{}])[0].get("code") == "NOT_FOUND":
                        # item wasn't found in marketplace, ignore error and return None
                        return
                except (ValueError, KeyError, IndexError):
                    pass  # If we can't parse the response as JSON, continue to raise the original error

                raise Exception(f"Error {response.status_code}: {response.text}")

        # If exhausted all retries
        raise Exception(
            f"Max retries ({max_retries}) exceeded for rate limit. Last error: {response.text}"
        )



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator
from ewah.hooks.amazon_ads import EWAHAmazonAdsHook

from datetime import timedelta


class EWAHAmazonAdsOperator(EWAHBaseOperator):
    _NAMES = ["amazon_ads"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: False,
    }

    def __init__(
        self,
        ads_type,
        report_type,
        profile_id=None,  # Pulls all available profiles if none is given
        additional_params=None,
        api_version="v2",  # default as long as not all endpoints migrated
        *args,
        **kwargs,
    ):
        # TODO: make error more explicit
        # TODO: validate via hook
        assert (ads_type in ("sp", "sb", "sd", "hsa") and api_version == "v2") or (
            ads_type in ("SPONSORED_PRODUCTS", "SPONSORED_BRANDS", "SPONSORED_DISPLAY") and api_version == "v3"
        ), "ads_type and api_version combination not supported"

        assert api_version in ("v2", "v3"), "only Amazon Ads v2 or v3 supported"

        if isinstance(profile_id, int):
            profile_id = str(profile_id)
        _msg = "profile_id, if supplied, must be a string or integer!"
        assert profile_id is None or isinstance(profile_id, str), _msg

        # TODO: auto-generate primary key (combined date + id, probably)

        super().__init__(*args, **kwargs)

        self.profile_id = profile_id
        self.ads_type = ads_type
        self.report_type = report_type
        self.api_version = api_version
        self.additional_params = additional_params

    def ewah_execute(self, context):
        # Note: Incremental only!
        start_date = self.data_from.date()
        end_date = self.data_until.date()
        if isinstance(self.profile_id, str):
            profiles = [{"profileId": self.profile_id}]
        elif self.profile_id is None:
            profiles = self.source_hook.get_profile_ids()
        else:
            raise Exception("This should not happen")

        if self.api_version == "v2":
            while start_date <= end_date:
                self.log.info(f"Requesting data for {start_date.isoformat()}...")
                # Get data for each profile
                for profile in profiles:
                    if profile.get("accountInfo", {}).get("type") == "agency":
                        continue
                    profile_id = profile["profileId"]
                    self.log.info(f"Requesting data for profile {profile_id}...")
                    
                    report_data = self.source_hook.get_report(
                        date=start_date,
                        profile_id=profile_id,
                        ads_type=self.ads_type,
                        report_type=self.report_type,
                        additional_params=self.additional_params,
                    )

                    # Enrich all records with metadata
                    for datum in report_data:
                        datum["_report_date"] = start_date.isoformat()
                        datum["_profile_id"] = profile_id
                        datum["_profile_metadata"] = profile
                    
                    # Upload once after enriching all records
                    self.upload_data(report_data)
                start_date += timedelta(days=1)
                
        if self.api_version == "v3":
            # Get data for each profile
            for profile in profiles:
                if profile.get("accountInfo", {}).get("type") == "agency":
                    continue
                profile_id = profile["profileId"]
                self.log.info(f"Requesting data for profile {profile_id}...")
                self.log.info(f"Start date: {start_date.isoformat()}")
                self.log.info(f"End date: {end_date.isoformat()}")
                
                report_data = self.source_hook.get_report_v3(
                    start_date=start_date,
                    end_date=end_date,
                    profile_id=profile_id,
                    ad_product=self.ads_type,
                    report_type=self.report_type,
                    additional_params=self.additional_params,
                )

                # Enrich all records with metadata
                for datum in report_data:
                    datum["_report_date"] = end_date.isoformat()
                    datum["_profile_id"] = profile_id
                    datum["_profile_metadata"] = profile
                
                # Upload once after enriching all records
                self.upload_data(report_data)



class EWAHAmazonAdsDSPOperator(EWAHBaseOperator):
    # Specifically for DSP endpoints

    _NAMES = ["amazon_ads_dsp"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: False,
    }

    def __init__(
        self,
        report_type,
        account_id,
        metrics=None,
        dimensions=None,
        request_period_days=30,
        *args,
        **kwargs,
    ):
        # TODO: auto-generate primary key (combined date + id, probably)

        super().__init__(*args, **kwargs)

        self.account_id = account_id
        self.report_type = report_type
        self.metrics = metrics
        self.dimensions = dimensions
        self.request_period_days = request_period_days

    def ewah_execute(self, context):
        # Note: Incremental only!
        request_date_start = self.data_from.date()
        request_date_end = min(
            self.data_until.date(),
            request_date_start + timedelta(days=self.request_period_days),
        )

        while True:
            # Maximum 30 days range per request
            self.log.info(
                f"Requesting data from {request_date_start.isoformat()} "
                f"until {request_date_end.isoformat()}..."
            )
            report_data = self.source_hook.get_report_dsp(
                account_id=self.account_id,
                report_date_start=request_date_start.isoformat(),
                report_date_end=request_date_end.isoformat(),
                report_type=self.report_type,
                metrics=self.metrics,
                dimensions=self.dimensions,
            )
            self.upload_data(report_data)
            if request_date_end == self.data_until.date():
                break
            else:
                request_date_start = request_date_end + timedelta(days=1)
                request_date_end = min(
                    self.data_until.date(),
                    request_date_start + timedelta(days=self.request_period_days),
                )



from ewah.hooks.base import EWAHBaseHook

import requests
import json
import time
from dateutil.parser import parse
from pytz import UTC


class EWAHShopifyGraphQLHook(EWAHBaseHook):
    _ATTR_RELABEL = {}

    # Note: so far only query implemenation for orders node
    # with line Items as JSON column

    conn_name_attr = "ewah_shopify_graphql_conn_id"
    default_conn_name = "ewah_shopify_graphql_default"
    conn_type = "ewah_shopify_graphql"
    hook_name = "EWAH Shopify GraphQL Connection"
    DEFAULT_API_VERSION = "2025-10"
    _BASE_URL = "https://{shop}.myshopify.com/admin/api/{version}/graphql.json"

    _ENDPOINTS = [
        "orders",
    ]

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "host", "port", "schema"],
            "relabeling": {
                "login": "Shop Subdomain (https://{this_part}.myshopify.com)",
                "password": "Access Token",
            },
        }

    def execute_graphql_query(
        self, query, variables=None, shop=None, version=None, max_retries=3
    ):
        """Execute a GraphQL query against Shopify's GraphQL API with retry logic"""
        headers = {
            "X-Shopify-Access-Token": self.conn.password,
            "Content-Type": "application/json",
            "X-GraphQL-Cost-Include-Fields": "true",
        }

        payload = {"query": query}

        shop = shop or self.conn.login
        version = version or self.DEFAULT_API_VERSION

        if variables:
            payload["variables"] = variables

        retry_delay = 1.0
        for attempt in range(max_retries + 1):
            try:
                response = requests.post(
                    self._BASE_URL.format(
                        shop=shop,
                        version=version,
                    ),
                    headers=headers,
                    json=payload,
                )
                response.raise_for_status()

                data = response.json()

                # Log cost information if available:
                # (1) requestedQueryCost: estimated before execution
                # (2) actualQueryCost: real cost after execution (max 1,000 points allowed)
                #      Note: "Calls to the GraphQL Admin API are limited based on calculated query
                #      costs, which means you should consider the cost of requests over time,
                #      rather than the number of requests.
                # (3) currentlyAvailable: leaky bucket capacity,
                #      see https://shopify.dev/docs/api/usage/limits#the-leaky-bucket-algorithm
                if "extensions" in data and "cost" in data["extensions"]:
                    cost = data["extensions"]["cost"]
                    self.log.info(
                        f"Query cost - Requested: {cost.get('requestedQueryCost')}, "
                        f"Actual: {cost.get('actualQueryCost')}, "
                        f"Available: {cost.get('throttleStatus', {}).get('currentlyAvailable')}"
                    )

                if "errors" in data:
                    raise Exception(f"GraphQL Errors: {data['errors']}")

                return data.get("data")

            except requests.exceptions.RequestException as e:
                if attempt == max_retries:
                    raise e
                self.log.warning(
                    f"Request failed (attempt {attempt + 1}/{max_retries + 1}): {e}"
                )
                time.sleep(retry_delay * (2**attempt))

    def flatten_order(self, order_node):

        flattened = {}

        # Basic order fields
        flattened["id"] = order_node.get("id", "").split("/")[-1]  # Extract ID from GID
        flattened["legacyResourceId"] = order_node.get("legacyResourceId")
        flattened["gid"] = order_node.get("id")
        flattened["name"] = order_node.get("name")
        flattened["number"] = order_node.get("number")
        flattened["email"] = order_node.get("email")
        flattened["createdAt"] = order_node.get("createdAt")
        flattened["updatedAt"] = order_node.get("updatedAt")
        flattened["processedAt"] = order_node.get("processedAt")
        flattened["cancelledAt"] = order_node.get("cancelledAt")
        flattened["cancelReason"] = order_node.get("cancelReason")
        flattened["closedAt"] = order_node.get("closedAt")
        flattened["currencyCode"] = order_node.get("currencyCode")
        flattened["customerLocale"] = order_node.get("customerLocale")
        flattened["note"] = order_node.get("note")
        flattened["tags"] = order_node.get("tags")
        flattened["totalWeight"] = order_node.get("totalWeight")
        flattened["displayFinancialStatus"] = order_node.get("displayFinancialStatus")
        flattened["displayFulfillmentStatus"] = order_node.get(
            "displayFulfillmentStatus"
        )
        flattened["sourceName "] = order_node.get("sourceName ")
        flattened["returnStatus"] = order_node.get("returnStatus")
        flattened["statusPageUrl"] = order_node.get("statusPageUrl")
        flattened["taxesIncluded"] = order_node.get("taxesIncluded")
        flattened["currentSubtotalPriceSet"] = order_node.get("currentSubtotalPriceSet")
        flattened["totalPriceSet"] = order_node.get("totalPriceSet")
        flattened["totalTaxSet"] = order_node.get("totalTaxSet")
        flattened["totalDiscountsSet"] = order_node.get("totalDiscountsSet")
        flattened["discountCodes"] = order_node.get("discountCodes")

        # discountApplications -> extract the node values
        disc_app_edges = order_node.get("discountApplications", {}).get("edges", [])
        disc_apps = [edge.get("node", {}) for edge in disc_app_edges]
        flattened["discountApplications"] = disc_apps

        # customer - extract (no egdes)
        customer_node = order_node.get("customer")
        flattened["customer"] = json.dumps(customer_node) if customer_node else None

        # lineItems -> extract the node values
        line_item_edges = order_node.get("lineItems", {}).get("edges", [])
        line_items = [edge.get("node", {}) for edge in line_item_edges]
        flattened["lineItems"] = line_items

        return flattened

    def get_data(
        self,
        endpoint,
        shop_id=None,
        version=None,
        first=50,
        data_from=None,
    ):
        """Get data from Shopify GraphQL API with pagination
        
        Note on pagination limits:
        - Orders: max 250 per page (controlled by 'first' parameter, default 50)
        - Line items: max 250 per order (hardcoded in query)
        - Discount applications: max 250 per order (hardcoded in query)
        
        For cases where we expect more items than these limits (e.g., orders with >250 line items,
        or orders with >250 discount applications), additional pagination logic would be needed
        to fetch subsequent pages using cursor-based pagination.
        """
        assert (
            endpoint in self._ENDPOINTS
        ), f"Invalid endpoint '{endpoint}'! Valid endpoints: {', '.join(self._ENDPOINTS)}"
        # Note: Currently only "orders" endpoint is implemented
        assert (
            endpoint == "orders"
        ), f"Endpoint '{endpoint}' not yet implemented. Only 'orders' is supported."

        shop_id = shop_id or self.conn.login
        version = version or self.DEFAULT_API_VERSION

        query = """
        query getOrders($first: Int!, $after: String, $query: String) {
            orders(first: $first, after: $after, query: $query) {
                pageInfo {
                    hasNextPage
                    endCursor
                }
                edges {
                    node {
                        id
                        legacyResourceId
                        name
                        number
                        email
                        createdAt
                        updatedAt
                        processedAt
                        closedAt
                        cancelledAt
                        cancelReason
                        currencyCode
                        customer {
                            id
                            emailMarketingConsent {
                                marketingState
                                marketingOptInLevel
                                consentUpdatedAt
                            }
                        }
                        customerLocale
                        taxesIncluded
                        currentSubtotalPriceSet { shopMoney { amount currencyCode } }
                        totalPriceSet { shopMoney { amount currencyCode } }
                        totalTaxSet { shopMoney { amount currencyCode } }
                        totalDiscountsSet { shopMoney { amount currencyCode } }
                        discountCodes
                        discountApplications(first: 250) {
                            edges {
                                node {
                                __typename
                                index
                                targetSelection
                                targetType
                                allocationMethod

                                ... on DiscountCodeApplication {
                                    code
                                }
                                ... on AutomaticDiscountApplication { title }
                                ... on ManualDiscountApplication { title description }
                                ... on ScriptDiscountApplication { title }

                                value {
                                    __typename
                                    ... on MoneyV2 { amount currencyCode }
                                    ... on PricingPercentageValue { percentage }
                                }
                                }
                            }
                            }

                        displayFinancialStatus
                        displayFulfillmentStatus
                        note
                        tags
                        totalWeight
                        statusPageUrl
                        returnStatus
                        sourceName

                        lineItems(first: 250) {
                            edges {
                                node {
                                    id
                                    name
                                    title
                                    quantity
                                    refundableQuantity
                                    sku
                                    customAttributes { key value }
                                    totalDiscountSet { shopMoney { amount currencyCode }}
                                    originalUnitPriceSet { shopMoney { amount currencyCode }}
                                    taxLines {
                                        title
                                        rate
                                        priceSet {
                                            shopMoney { amount currencyCode }
                                        }
                                    }
                                    discountAllocations {
                                        allocatedAmountSet { shopMoney { amount currencyCode } }
                                        discountApplication { index }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        """

        has_next_page = True
        cursor = None

        # Build query string for GraphQL if data_from provided
        query_string = None
        if data_from:
            # ensure UTC timezone and ISO 8601 format
            dt = parse(str(data_from))
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=UTC)

            iso_string = dt.isoformat().replace("+00:00", "Z")
            query_string = f"updated_at:>='{iso_string}'"
            self.log.info(f"Filtering orders with query: {query_string}")

        while has_next_page:
            variables = {"first": first}
            if cursor:
                variables["after"] = cursor
            if query_string:
                variables["query"] = query_string

            self.log.info(f"Fetching orders with cursor: {cursor}")

            data = self.execute_graphql_query(
                query, variables, shop=shop_id, version=version
            )

            if not data or "orders" not in data:
                break

            orders_data = data["orders"]
            page_info = orders_data["pageInfo"]
            orders = orders_data["edges"]

            # Flatten and yield batch
            flattened_orders = []
            for edge in orders:
                flattened_order = self.flatten_order(edge["node"])
                flattened_orders.append(flattened_order)

            if flattened_orders:
                yield flattened_orders

            has_next_page = page_info.get("hasNextPage", False)
            cursor = page_info.get("endCursor")



from ewah.hooks.sql_base import EWAHSQLBaseHook
from ewah.constants import EWAHConstants as EC

from google.cloud import bigquery
from google.oauth2 import service_account
from tempfile import NamedTemporaryFile
from datetime import timedelta

from typing import List, Dict, Union, Type, Optional

import os
import json


class EWAHBigQueryHook(EWAHSQLBaseHook):
    _ATTR_RELABEL: dict = {"project": "host"}

    conn_name_attr = "ewah_bigquery_conn_id"
    default_conn_name = "ewah_bigquery_default"
    conn_type = "ewah_bigquery"
    hook_name = "EWAH BigQuery Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "password", "login", "schema", "port"],
            "relabeling": {"host": "Project ID"},
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        # from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from ewah.utils.widgets import EWAHTextAreaWidget
        from wtforms import StringField

        return {
            "extra__ewah_bigquery__service_account_json": StringField(
                "Service Account JSON",
                widget=EWAHTextAreaWidget(rows=12),
            ),
            "extra__ewah_bigquery__location": StringField(
                "[optional] location",
                widget=EWAHTextAreaWidget(rows=1),
            ),
        }

    def __init__(self, *args, project_id=None, **kwargs):
        self.project_id = project_id
        return super().__init__(*args, **kwargs)

    class bq_cursor(object):
        """Wrapper around the BigQuery client to work similarly to other db cursor.

        :param client: An initialized and authenticated bigquery Client object.
        :param return_dict: Set to true to return lists of dictionaries in fetach calls,
            otherwise fetches return lists of lists. Defaults to True.
        """

        def __init__(
            self,
            outer_class,
            client: Type[bigquery.client.Client],
            return_dict: bool = True,
        ):
            self.client = client
            self.outer = outer_class  # access outer class
            self.return_dict = return_dict
            self.latest_query = None

        def close(self):
            self.outer.log.info("Warning: BigQuery cursors cannot be closed!")

        def execute(self, sql, params=None, commit=True):
            if not commit:
                self.outer.log.info("Warning: BigQuery always auto-commits!")

            # params have to be supplied via a job_config object!
            # only scalar implemented - array and structs are TODO!
            if params:
                type_mapping = EC.QBC_TYPE_MAPPING.get(EC.DWH_ENGINE_BIGQUERY)
                _default = type_mapping[str]
                job_config = bigquery.QueryJobConfig(
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            key,
                            type_mapping.get(type(value), _default),
                            value,
                        )
                        for key, value in params.items()
                    ]
                )
            else:
                job_config = None

            self.latest_query = self.client.query(sql, job_config=job_config)

        def fetchall(self):
            return self.fetch_all()

        def fetch_all(self):
            if not self.latest_query:
                return []

            dict_result = [dict(row) for row in self.latest_query.result()]
            if self.return_dict:
                return dict_result

            # Note: Must deal with empty results
            keys = (dict_result[0] if dict_result else {}).keys()
            return [[row[key] for key in keys] for row in dict_result]

        def fetch_in_batches(self, batch_size: int):
            """Generator to iterate over pages of a custom size."""
            if not self.latest_query:
                return []
            for page in self.latest_query.result(page_size=batch_size).pages:
                self.outer.log.info("Yielding a page...")
                dict_result = [dict(row) for row in page]
                if self.return_dict:
                    yield dict_result
                else:
                    keys = dict_result[0].keys()
                    yield [[row[key] for key in keys] for row in dict_result]

    def _get_db_conn(self):
        conn_kwargs = {
            "credentials": service_account.Credentials.from_service_account_info(
                json.loads(self.conn.service_account_json)
            )
        }
        if self.conn.location:
            conn_kwargs["location"] = self.conn.location
        if self.project_id or self.conn.project:
            conn_kwargs["project"] = self.project_id or self.conn.project
        return bigquery.client.Client(**conn_kwargs)

    def _get_cursor(self):
        return self.bq_cursor(outer_class=self, client=self.dbconn, return_dict=False)

    def _get_dictcursor(self):
        return self.bq_cursor(outer_class=self, client=self.dbconn, return_dict=True)

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(
            "Executing SQL:\n\n{0}\n\nWith params:\n{1}".format(
                sql,
                "\n".join(
                    [
                        "{0}: {1}".format(key, str(value))
                        for (key, value) in params.items()
                    ]
                )
                if params
                else "No params!",
            )
        )
        (cursor or self.cursor).execute(sql=sql.strip(), params=params, commit=commit)

    def get_data_from_sql(
        self, sql: str, params: Optional[dict] = None, return_dict: bool = True
    ) -> Union[List[list], List[dict]]:
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur, commit=False)
        return cur.fetchall()

    def get_data_in_batches(
        self,
        sql: str,
        params: Optional[dict] = None,
        return_dict: bool = True,
        batch_size: int = 25000,
    ):
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur)
        data = []
        for batch in cur.fetch_in_batches(batch_size=batch_size):
            data += batch
            if len(data) >= batch_size:
                yield data
                data = []
        if data:
            yield data



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.hooks.bigquery import EWAHBigQueryHook


class EWAHBigQueryOperator(EWAHSQLBaseOperator):
    _NAMES = ["bq", "biqguery"]

    _SQL_BASE = """
        SELECT
            {columns}
        FROM `{database}`.`{schema}`.`{table}`
    """
    _SQL_BASE_SELECT = "SELECT * FROM ({select_sql}) t \nWHERE {{0}}"
    _SQL_COLUMN_QUOTE = "`"
    _SQL_PARAMS = "@{0}"

    _CONN_TYPE = EWAHBigQueryHook.conn_type

    def __init__(self, *args, **kwargs):
        _msg = "Must supply source_database_name (=project_id )!"
        if kwargs.get("project_id"):
            kwargs["source_database_name"] = kwargs.pop("project_id")
        assert kwargs.get("source_database_name"), _msg
        super().__init__(*args, **kwargs)



from ewah.constants import EWAHConstants as EC
from ewah.hooks.shopify_graphql import EWAHShopifyGraphQLHook
from ewah.operators.base import EWAHBaseOperator


class EWAHShopifyGraphQLOperator(EWAHBaseOperator):
    _NAMES = ["shopify_graphql"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_SUBSEQUENT: True,
    }

    def __init__(
        self,
        endpoint=None,
        shop_id=None,
        api_version=None,
        first=250,  # 250 is the max
        *args,
        **kwargs,
    ):
        self.endpoint = (endpoint or kwargs.get("target_table_name", "")).lower()

        assert (
            self.endpoint in EWAHShopifyGraphQLHook._ENDPOINTS
        ), f"Invalid endpoint '{self.endpoint}'! Valid endpoints: {', '.join(EWAHShopifyGraphQLHook._ENDPOINTS)}"

        # Set default subsequent_field for orders
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get("subsequent_field", "updatedAt")

        # Set default primary key
        kwargs["primary_key"] = kwargs.get(
            "primary_key", "gid"
        )  # ex: gid://shopify/{object_name}/{id}
        super().__init__(*args, **kwargs)
        self.shop_id = shop_id
        self.api_version = api_version
        self.first = first

    def ewah_execute(self, context):
        self._metadata.update({"shop_id": self.shop_id})

        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            self.log.info(
                f"""Found max timestamp in result table -> Subsequent load:
                fetching orders updated after {data_from}"""
            )
        else:
            data_from = self.data_from

        for batch in self.source_hook.get_data(
            endpoint=self.endpoint,
            shop_id=self.shop_id,
            version=self.api_version,
            first=self.first,
            data_from=data_from,
        ):
            self.upload_data(batch)
